import fmpy
import zipfile
import time
from ..util import *


def cross_check(fmus_dir, report, result_dir, simulate, tool_name, tool_version, skip, readme):

    html = open(report, 'w')
    html.write('''<html>
    <head>
        <style>
            p, li, td, th, footer {
              font-family: 'Helvetica Neue', Helvetica, Arial, sans-serif;
              font-size: 0.75em;
              line-height: 1.5em;
            }
            td.passed { background-color: #eeffee; }
            td.failed { background-color: #ffdddd; }

            td.status { text-align: center; }

            .label {
                display: inline-block;
                min-width: 35px;
                padding: .5em .6em .5em;
                font-size: 75%;
                font-weight: 700;
                line-height: 1;
                color: #fff;
                text-align: center;
                white-space: nowrap;
                vertical-align: baseline;
                border-radius: .25em;
            }

            .label-danger {
                background-color: #d9534f;
            }

            .label-warning {
                background-color: #f0ad4e;
            }

            .label-success {
                background-color: #5cb85c;
            }

            .label-default {
                background-color: #777;
            }

            .tooltip {
                position: relative;
                display: inline-block;
                border-bottom: 1px dotted black;
            }

            .tooltip .tooltiptext {
                visibility: hidden;
                width: auto;
                background-color: black;
                color: #fff;
                text-align: center;
                border-radius: 6px;
                padding: 5px;

                /* Position the tooltip */
                position: absolute;
                z-index: 1;
                right: 0;
                top: 30px;
            }

            .tooltip:hover .tooltiptext {
                visibility: visible;
            }
        </style>
    </head>
    <body>
        <table>''')
    html.write('<tr><th>Model</th><th>FMU</th><th>Opts</th><th>In</th><th>Ref</th><th>Doc</th><th>Sim</th><th>Res</th></tr>\n')

    for root, dirs, files in os.walk(fmus_dir):

        fmu_filename = None

        if 'notCompliantWithLatestRules' in files:
            continue  # skip

        for file in files:
            if file.endswith('.fmu'):
                fmu_name, _ = os.path.splitext(file)  # FMU name without file extension
                fmu_filename = os.path.join(root, file)  # absolute path filename
                break

        if fmu_filename is None:
            continue  # skip

        # dictionary that contains the information about the FMU
        options = {'fmu_filename': fmu_filename}

        # extract the cross-check info from the path
        options.update(fmu_path_info(root))

        # skip FMUs in _FMIModelicaTest and other directories
        if options['fmi_version'] not in ['1.0', '2.0']:
            continue

        if skip(options):
            continue

        print(root)

        ##########################
        # VALIDATE FMU AND FILES #
        ##########################

        # read the model description
        try:
            model_description = fmpy.read_model_description(fmu_filename, validate=True, validate_variable_names=True)
            xml_cell = '<td class="status"><span class="label label-success">valid</span></td>'
        except Exception as e:
            # try again without validation
            model_description = fmpy.read_model_description(fmu_filename, validate=False)
            xml_cell = '<td class="status"><span class="label label-danger" title="' + str(e) + '">invalid</span></td>'

        # check for documentation
        if model_description.fmiVersion == '1.0':
            doc_path = 'documentation/_main.html'
        else:
            doc_path = 'documentation/index.html'

        doc_cell = '<td class="status"><span class="label label-default" title="' + doc_path + " is missing" + '">n/a</span></td>'

        with zipfile.ZipFile(fmu_filename, 'r') as zf:
            if doc_path in zf.namelist():
                # TODO: validate HTML?
                doc_cell = '<td class="status"><span class="label label-success">available</span></td>'

        input_variables = []
        output_variables = []

        # collect the variable names
        for variable in model_description.modelVariables:
            if variable.causality == 'input':
                input_variables.append(variable.name)
            elif variable.causality == 'output':
                output_variables.append(variable.name)

        # check the reference options file
        try:
            ref_opts = read_ref_opt_file(os.path.join(root, fmu_name + '_ref.opt'))
            ref_opts_cell = '<td class="status"><span class="label label-success">valid</span></td>'
        except Exception as e:
            ref_opts = None
            ref_opts_cell = '<td class="status"><span class="label label-danger" title="' + str(e) + '">invalid</span></td>'

        def check_csv_file(filename, variables):
            """ Load the CSV and generate the table cell """

            try:
                # load with validation
                traj = read_csv(filename=filename, variable_names=variables)
                cell = '<td class="status"><span class="label '
                cell += 'label-warning' if traj.size > 1000 else 'label-success'
                cell += '">%d &times; %d</span></td>' % (len(traj.dtype), traj.size)
            except Exception as e1:
                try:
                    # load without validation
                    traj = read_csv(filename=filename)
                    cell = '<td class="status"><span class="label label-warning" title="' + str(e1) + '">%d &times; %d</span></td>' % (len(traj.dtype), traj.size)
                except Exception as e2:
                    traj = None
                    cell = '<td class="status"><span class="label label-danger" title="' + str(e2) + '">invalid</span></td>'

            return traj, cell

        # check the input file
        input = None

        if input_variables:
            in_path = os.path.join(root, fmu_name + '_in.csv')
            input, in_csv_cell = check_csv_file(filename=in_path, variables=input_variables)
        else:
            in_path = None
            in_csv_cell = '<td class="status"><span class="label label-default">n/a</span></td>'

        # check the reference file
        ref_path = os.path.join(root, fmu_name + '_ref.csv')
        reference, ref_csv_cell = check_csv_file(filename=ref_path, variables=output_variables)

        # supported_platforms = fmpy.supported_platforms(fmu_filename)

        # this will remove any trailing (back)slashes
        fmus_dir = os.path.normpath(fmus_dir)
        model_path = fmu_filename[len(fmus_dir) + 1:]
        model_path = os.path.dirname(model_path)
        fmu_simple_filename = os.path.basename(fmu_filename)
        model_name, _ = os.path.splitext(fmu_simple_filename)

        # build the filenames
        result = None

        ##############
        # SIMULATION #
        ##############

        skipped = True

        if simulate is None:
            sim_cell = '<td class="status" title="Simulation is disabled"><span class="label label-default">skipped</span></td>'
        elif ref_opts is None:
            sim_cell = '<td class="status" title="Failed to read *_ref.opt file"><span class="label label-default">skipped</span></td>'
        elif input_variables and input is None:
            sim_cell = '<td class="status" title="Input file is invalid"><span class="label label-default">skipped</span></td>'
        else:
            skipped = False

            step_size = ref_opts['StepSize']
            stop_time = ref_opts['StopTime']

            if reference is not None:
                output_variable_names = reference.dtype.names[1:]
            else:
                output_variable_names = None

            try:
                start_time = time.time()

                options['fmu_filename'] = fmu_filename
                options['step_size'] = None if step_size == 0 else step_size
                options['stop_time'] = stop_time

                if in_path is not None:
                    options['input_filename'] = in_path

                options['output_variable_names'] = output_variable_names

                # simulate the FMU
                result = simulate(options)

                sim_cell = '<td class="status"><span class="label label-success">%.2f s</span></td>' % (time.time() - start_time)

            except Exception as e:
                sim_cell = '<td class="status"><span class="label label-danger" title="' + str(e) + '">failed</span></td>'

        ##############
        # VALIDATION #
        ##############
        if skipped:
            res_cell = '<span class="label label-default">n/a</span>'
        else:
            try:
                rel_out = validate_result(result=result, reference=reference)
            except Exception as e:
                print(str(e))
                rel_out = 1

            res_cell = '<span class="label '

            if rel_out <= 0:
                res_cell += 'label-success'
            elif rel_out <= 0.1:
                res_cell += 'label-warning'
            else:
                res_cell += 'label-danger'

            res_cell += '">%d %%</span>' % np.floor((1.0 - rel_out) * 100)

        # this will remove any trailing (back)slashes
        fmus_dir = os.path.normpath(fmus_dir)

        model_path = fmu_filename[len(fmus_dir)+1:]

        model_path = os.path.dirname(model_path)

        html.write('<tr><td>' + root + '</td>')
        html.write('\n'.join([xml_cell, ref_opts_cell, in_csv_cell, ref_csv_cell, doc_cell, sim_cell]))

        if result_dir is not None and result is not None:

            relative_result_dir = os.path.join(result_dir, options['fmi_version'], options['fmi_type'],
                                               options['platform'], tool_name, tool_version, options['tool_name'],
                                               options['tool_version'], options['model_name'])

            fmu_result_dir = os.path.join(result_dir, relative_result_dir)

            if not os.path.exists(fmu_result_dir):
                os.makedirs(fmu_result_dir)

            # write the indicator file
            if skipped:
                indicator_filename = 'rejected'
            elif rel_out < 0.1:
                indicator_filename = 'passed'
            else:
                indicator_filename = 'failed'

            with open(os.path.join(fmu_result_dir, indicator_filename), 'w') as f:
                pass

            if readme is not None:
                # write the ReadMe.txt file
                with open(os.path.join(fmu_result_dir, 'ReadMe.txt'), 'w') as f:
                    f.write(readme())

            result_filename = os.path.join(fmu_result_dir, model_name + '_out.csv')

            header = ','.join(map(lambda s: '"' + s + '"', result.dtype.names))
            np.savetxt(result_filename, result, delimiter=',', header=header, comments='', fmt='%g')

            reference_filename = os.path.join(fmus_dir, model_path, model_name + '_ref.csv')

            if os.path.isfile(reference_filename):
                # load the reference trajectories
                reference = np.genfromtxt(reference_filename, delimiter=',', names=True, deletechars='')
            else:
                reference = None

            plot_filename = os.path.join(fmu_result_dir, 'result.png')
            plot_result(result, reference, filename=plot_filename)

            html.write(r'<td><div class="tooltip">' + res_cell + '<span class="tooltiptext"><img src="'
                       + os.path.join(relative_result_dir, 'result.png').replace('\\', '/') + '"/></span ></div></td>')
        else:
            html.write('<td class="status">' + res_cell + '</td>\n')

        html.write('</tr>\n')

    html.write('</table></body></html>')

    print("Done.")



from PyQt5.QtCore import QThread, pyqtSignal, QDateTime
from fmpy import simulate_fmu
from fmpy.fmi2 import fmi2Warning


class SimulationThread(QThread):

    progressChanged = pyqtSignal(int)
    messageChanged = pyqtSignal(str, str)

    def __init__(self, filename, fmiType, stopTime, solver, stepSize, relativeTolerance, outputInterval, startValues, applyDefaultStartValues, input, output, debugLogging, fmiLogging, parent=None):

        super(SimulationThread, self).__init__(parent)

        self.filename = filename
        self.fmiType = fmiType
        self.stopTime = stopTime
        self.solver = solver
        self.stepSize = stepSize
        self.relativeTolerance = relativeTolerance
        self.outputInterval = outputInterval
        self.startValues = startValues
        self.applyDefaultStartValues = applyDefaultStartValues
        self.input = input
        self.output = output
        self.debugLogging = debugLogging
        self.fmiLogging = fmiLogging
        self.progress = 0
        self.stopped = False

        self.cols = []
        self.rows = []
        self.result = None

    def stop(self):
        self.stopped = True

    def logFMUMessage(self, componentEnvironment, instanceName, status, category, message):

        if status == fmi2Warning:
            level = 'warning'
        elif status > fmi2Warning:
            level = 'error'
        else:
            level = 'info'

        self.messageChanged.emit(level, message.decode('utf-8'))

    def logFMICall(self, message):
        self.messageChanged.emit('debug', message)

    def stepFinished(self, time, recorder):

        progress = int((time / self.stopTime) * 100)

        if progress > self.progress:
            self.progress = progress
            self.progressChanged.emit(progress)

        # self.msleep(10)

        self.cols = recorder.cols
        self.rows = recorder.rows

        return not self.stopped

    def run(self):

        startTime = QDateTime.currentMSecsSinceEpoch()

        try:
            self.result = simulate_fmu(self.filename,
                                       stop_time=self.stopTime,
                                       solver=self.solver,
                                       step_size=self.stepSize,
                                       relative_tolerance=self.relativeTolerance,
                                       output_interval=self.outputInterval,
                                       fmi_type=self.fmiType,
                                       start_values=self.startValues,
                                       apply_default_start_values=self.applyDefaultStartValues,
                                       input=self.input,
                                       output=self.output,
                                       debug_logging=self.debugLogging,
                                       visible=True,
                                       fmi_call_logger=self.logFMICall if self.fmiLogging else None,
                                       logger=self.logFMUMessage,
                                       step_finished=self.stepFinished)
        except Exception as e:
            self.messageChanged.emit('error', str(e))

        stopTime = QDateTime.currentMSecsSinceEpoch()
        totalTime = ((stopTime - startTime) / 1000.)

        if self.stopped:
            self.messageChanged.emit('info', 'Simulation stopped after %s s' % totalTime)
        else:
            self.messageChanged.emit('info', 'Simulation took %s s' % totalTime)

    def logFMICall(self, message):
        self.messageChanged.emit('debug', message)



""" This package contains the graphical user interface based on PyQt5 """


def compile_resources():
    """ Compile the .ui and .qrc files if they are available and newer than the compiled .py files """

    import os

    gui_dir = os.path.dirname(__file__)

    forms_dir = os.path.join(gui_dir, 'forms')
    generated_dir = os.path.join(gui_dir, 'generated')

    # compile the forms
    if os.path.isdir(forms_dir):

        for file in os.listdir(forms_dir):

            if not file.endswith('.ui'):
                continue

            form = os.path.basename(file)
            form, _ = os.path.splitext(form)

            ui_file = os.path.join(forms_dir, form + '.ui')
            py_file = os.path.join(generated_dir, form + '.py')

            if os.path.isfile(ui_file):
                if not os.path.isfile(py_file) or os.path.getmtime(ui_file) > os.path.getmtime(py_file):
                    print("UIC'ing %s" % ui_file)
                    os.system('pyuic5 %s -o %s --import-from .' % (ui_file, py_file))

    icons_qrc = os.path.join(gui_dir, 'icons', 'icons.qrc')
    icons_rc_py = os.path.join(generated_dir, 'icons_rc.py')

    # compile the resources
    if os.path.isfile(icons_qrc):
        if not os.path.isfile(icons_rc_py) or os.path.getmtime(icons_qrc) > os.path.getmtime(icons_rc_py):
            print("RCC'ing %s" % icons_qrc)
            os.system('pyrcc5 %s -o %s' % (icons_qrc, icons_rc_py))



from PyQt5.QtCore import Qt, QAbstractTableModel, QModelIndex
from PyQt5.QtGui import QFont, QIcon
from PyQt5.QtWidgets import QDialog, QHeaderView
import pyqtgraph as pg
import numpy as np

from fmpy.gui.generated.TableDialog import Ui_TableDialog

pg.setConfigOptions(background='w', foreground='k', antialias=True)


class TableModel(QAbstractTableModel):

    def __init__(self, size, data, startValues, parent=None):
        super(TableModel, self).__init__(parent)
        self.size = size
        self.data = data
        self.startValues = startValues
        self.boldFont = QFont()
        self.boldFont.setBold(True)

    def rowCount(self, index=QModelIndex()):
        return self.size[0]

    def columnCount(self, index=QModelIndex()):
        if len(self.size) == 1:
            return 1
        else:
            return self.size[1]

    def headerData(self, section, orientation, role):

        if role == Qt.DisplayRole:

            if orientation == Qt.Horizontal:
                if len(self.size) > 1:
                    return section + 1
            else:
                return section + 1

        return None

    def flags(self, index):
        flags = super(TableModel, self).flags(index) | Qt.ItemIsEditable
        return flags

    def data(self, index, role=Qt.DisplayRole):

        if not index.isValid():
            return None

        subs = self.ind2sub(index)
        sv = self.data[subs]

        if role == Qt.DisplayRole or role == Qt.EditRole:
            if sv.name in self.startValues:
                return self.startValues[sv.name]
            else:
                return sv.start

        elif role == Qt.TextAlignmentRole:
            return Qt.AlignRight | Qt.AlignVCenter

        elif role == Qt.FontRole and sv.name in self.startValues:
            return self.boldFont

        return None

    def setData(self, index, value, role):
        subs = self.ind2sub(index)
        sv = self.data[subs]

        if value:
            self.startValues[sv.name] = value
        else:
            self.startValues.pop(sv.name, None)

        self.dataChanged.emit(index, index)

        return True

    def ind2sub(self, index):
        """ Convert index to subscripts """

        if len(self.size) == 1:
            return index.row(),
        else:
            return index.row(), index.column()


class TableDialog(QDialog):
    """ A dialog to edit 1-d and 2-d array variables """

    def __init__(self, modelVariables, variable, startValues, parent=None):
        """
        Parameters:
            modelVariables: list of variables from the model description
            variable:       the (array) variable to be edited
            startValues:    dictionary of modified start values
        """

        super(TableDialog, self).__init__(parent)

        self.modelVariables = modelVariables
        self.startValues = startValues

        self.ui = Ui_TableDialog()
        self.ui.setupUi(self)

        # hide the question mark button
        flags = self.windowFlags()
        flags &= ~Qt.WindowContextHelpButtonHint
        self.setWindowFlags(flags)

        verticalHeader = self.ui.tableView.verticalHeader()
        verticalHeader.setSectionResizeMode(QHeaderView.Fixed)
        verticalHeader.setDefaultSectionSize(24)

        self.data = dict()
        size = None

        variable_name = variable.name
        i = variable_name.rfind('[')
        variable_name = variable_name[:i]

        for sv in self.modelVariables:
            if sv.name.startswith(variable_name):
                i = sv.name.rfind('[')
                subs = sv.name[i+1:-1]
                subs = subs.split(',')
                subs = list(map(lambda s: int(s) - 1, subs))
                if size is None:
                    size = subs
                else:
                    for i, index in enumerate(subs):
                        size[i] = max(size[i], index + 1)

                self.data[tuple(subs)] = sv

        self.size = size

        self.setWindowTitle(variable_name + '[' + ','.join(map(str, size)) + ']')
        self.setWindowIcon(QIcon(':/icons/%s.png' % variable.type.lower()))

        self.model = TableModel(size, self.data, self.startValues)
        self.ui.tableView.setModel(self.model)

        self.ui.tableView.resizeColumnsToContents()

        self.updatePlot()

        self.ui.splitter.setSizes([450, 450])

        if len(size) < 2:
            self.ui.plotSettingsWidget.hide()

        self.model.dataChanged.connect(self.updatePlot)

        self.ui.plotColumnsRadioButton.toggled.connect(self.updatePlot)
        self.ui.firstColumnAsXAxisCheckBox.toggled.connect(self.updatePlot)
        self.ui.firstRowAsXAxisCheckBox.toggled.connect(self.updatePlot)

    @staticmethod
    def canEdit(variable):
        """ Determine whether the variable can be edited """

        if variable.type not in ['Real', 'Integer', 'Enumeration']:
            return False

        try:
            subs = TableDialog.subscripts(variable.name)
            return 0 < len(subs) <= 2
        except:
            return False

        return False

    @staticmethod
    def subscripts(name):
        """ Get the subscripts from a variable name """

        i = name.rfind('[')
        subs = name[i + 1:-1]
        subs = subs.split(',')
        return list(map(lambda s: int(s) - 1, subs))

    def updatePlot(self):

        data = np.zeros(self.size)

        for subs, variable in self.data.items():
            if variable.name in self.startValues:
                data[subs] = float(self.startValues[variable.name])
            else:
                if variable.start:
                    data[subs] = float(variable.start)
                else:
                    data[subs] = np.nan

        self.ui.graphicsView.clear()
        plot = self.ui.graphicsView.addPlot()
        plot.showGrid(x=True, y=True, alpha=0.25)

        if len(self.size) == 1:
            plot.plot(y=data, pen='b', symbolPen=None, symbolBrush='b', symbolSize=5)
        else:
            vb = self.ui.graphicsView.addViewBox()
            vb.setMaximumWidth(100)
            legend = pg.LegendItem()
            legend.setParentItem(vb)
            # Anchor the upper-left corner of the legend to the upper-left corner of its parent
            legend.anchor((0, 0), (0, 0))

            if self.ui.plotColumnsRadioButton.isChecked():

                if self.ui.firstColumnAsXAxisCheckBox.isChecked():
                    for i in range(1, min(data.shape[1], 20)):
                        color = pg.intColor(i, hues=data.shape[1] - 1, maxValue=200)
                        line = plot.plot(x=data[:, 0], y=data[:, i], pen=color, symbolPen=None, symbolBrush=color, symbolSize=5)
                        legend.addItem(line, 'Column %d' % (i + 1))

                else:
                    for i in range(data.shape[1]):
                        color = pg.intColor(i, hues=data.shape[1], maxValue=200)
                        line = plot.plot(y=data[:, i], pen=color, symbolPen=None, symbolBrush=color, symbolSize=5)
                        legend.addItem(line, 'Column %d' % (i + 1))

            else:

                n = min(data.shape[0], 30)

                if self.ui.firstRowAsXAxisCheckBox.isChecked():

                    for i in range(1, n):
                        color = pg.intColor(i, hues=n, maxValue=200)
                        line = plot.plot(x=data[0, :], y=data[i, :], pen=color, symbolPen=None, symbolBrush=color, symbolSize=5)
                        legend.addItem(line, 'Row %d' % (i + 1))
                else:
                    for i in range(n):
                        color = pg.intColor(i, hues=n, maxValue=200)
                        line = plot.plot(y=data[i, :], pen=color, symbolPen=None, symbolBrush=color, symbolSize=5)
                        legend.addItem(line, 'Row %d' % (i + 1))



from tempfile import mkdtemp

from attr import attrs


@attrs(eq=False, auto_attribs=True)
class Variable(object):

    type: str = None
    variability: str = None
    causality: str = None
    name: str = None
    start: str = None
    description: str = None
    mapping: str = None


@attrs(eq=False, auto_attribs=True)
class Connection(object):

    startElement: str
    startConnector: str
    endElement: str
    endConnector: str


def create_fmu_container(configuration, output_filename):
    """ Create an FMU from nested FMUs (experimental)

        see tests/test_fmu_container.py for an example
    """

    def xml_encode(s):
        """ Escape non-ASCII characters """

        s = s.replace('&', '&amp;')
        s = s.replace('<', '&lt;')
        s = s.replace('>', '&gt;')
        s = s.replace('"', '&quot;')

        for c in s:
            if ord(c) > 127:
                s = s.replace(c, '&#x' + format(ord(c), 'x') + ';')

        return s

    import os
    import shutil
    import fmpy
    from fmpy import read_model_description, extract
    import msgpack
    from datetime import datetime
    import pytz

    base_filename, _ = os.path.splitext(output_filename)
    model_name = os.path.basename(base_filename)

    unzipdir = mkdtemp()

    basedir = os.path.dirname(__file__)

    for directory in ['binaries', 'documentation', 'sources']:
        shutil.copytree(os.path.join(basedir, directory), os.path.join(unzipdir, directory))

    os.mkdir(os.path.join(unzipdir, 'resources'))

    data = {
        'components': [],
        'variables': [],
        'connections': []
    }

    component_map = {}

    for i, component in enumerate(configuration['components']):
        model_description = read_model_description(component['filename'])
        model_identifier = model_description.coSimulation.modelIdentifier
        extract(component['filename'], os.path.join(unzipdir, 'resources', model_identifier))
        variables = dict((v.name, v) for v in model_description.modelVariables)
        component_map[component['name']] = (i, variables)
        data['components'].append({
            'name': component['name'],
            'guid': model_description.guid,
            'modelIdentifier': model_identifier,
        })

    variables_map = {}

    for i, v in enumerate(configuration['variables']):
        variables_map[v.name] = (i, v)

    mv = ''  # model variables
    mo = ''  # model outputs

    for i, v in enumerate(configuration['variables']):

        component_indices = []
        value_references = []

        # config.mp
        for component_name, variable_name in v.mapping:
            component_index, component_variables = component_map[component_name]
            value_reference = component_variables[variable_name].valueReference
            component_indices.append(component_index)
            value_references.append(value_reference)

        data['variables'].append({
            'components': component_indices,
            'valueReferences': value_references
        })

        # modelDescription.xml
        start = f' start="{ v.start }"' if v.start else ''
        mv += f'\n    <ScalarVariable name="{ v.name }" valueReference="{ i }" variability="{ v.variability }" causality="{ v.causality }" description="{ v.description }">'
        mv += f'\n      <{v.type}{ start }/>'
        mv += f'\n    </ScalarVariable>'

        if v.causality == 'output':
            mo += f'\n      <Unknown index="{ i + 1 }"/>'

    xml = f'''<?xml version="1.0" encoding="UTF-8"?>
<fmiModelDescription
  fmiVersion="2.0"
  modelName="{ model_name }"
  guid=""
  description="{ configuration.get('description', '') }"
  generationTool="FMPy {fmpy.__version__} FMU Container"
  generationDateAndTime="{ datetime.now(pytz.utc).isoformat() }">

  <CoSimulation modelIdentifier="FMUContainer">
    <SourceFiles>
      <File name="FMUContainer.c"/>
      <File name="mpack.c"/>
    </SourceFiles>
  </CoSimulation>

  <ModelVariables>{ mv }
  </ModelVariables>

  <ModelStructure>
    <Outputs>{mo}
    </Outputs>
    <InitialUnknowns>{ mo }
    </InitialUnknowns>
  </ModelStructure>

</fmiModelDescription>
'''

    with open(os.path.join(unzipdir, 'modelDescription.xml'), 'w') as f:
        f.write(xml)

    for c in configuration['connections']:
        data['connections'].append({
            'type': component_map[c.startElement][1][c.startConnector].type,
            'startComponent': component_map[c.startElement][0],
            'endComponent': component_map[c.endElement][0],
            'startValueReference': component_map[c.startElement][1][c.startConnector].valueReference,
            'endValueReference': component_map[c.endElement][1][c.endConnector].valueReference,
        })

    with open(os.path.join(unzipdir, 'resources', 'config.mp'), 'wb') as f:
        packed = msgpack.packb(data)
        f.write(packed)

    shutil.make_archive(base_filename, 'zip', unzipdir)

    if os.path.isfile(output_filename):
        os.remove(output_filename)

    os.rename(base_filename + '.zip', output_filename)

    shutil.rmtree(unzipdir, ignore_errors=True)



from .sundials_nvector import *
from .sundials_types import sunindextype, booleantype, realtype
from .libraries import sundials_nvecserial

# /* -----------------------------------------------------------------
#  * Programmer(s): Scott D. Cohen, Alan C. Hindmarsh, Radu Serban,
#  *                and Aaron Collier @ LLNL
#  * -----------------------------------------------------------------
#  * SUNDIALS Copyright Start
#  * Copyright (c) 2002-2019, Lawrence Livermore National Security
#  * and Southern Methodist University.
#  * All rights reserved.
#  *
#  * See the top-level LICENSE and NOTICE files for details.
#  *
#  * SPDX-License-Identifier: BSD-3-Clause
#  * SUNDIALS Copyright End
#  * -----------------------------------------------------------------
#  * This is the header file for the serial implementation of the
#  * NVECTOR module.
#  *
#  * Notes:
#  *
#  *   - The definition of the generic N_Vector structure can be found
#  *     in the header file sundials_nvector.h.
#  *
#  *   - The definition of the type 'realtype' can be found in the
#  *     header file sundials_types.h, and it may be changed (at the
#  *     configuration stage) according to the user's needs.
#  *     The sundials_types.h file also contains the definition
#  *     for the type 'booleantype'.
#  *
#  *   - N_Vector arguments to arithmetic vector operations need not
#  *     be distinct. For example, the following call:
#  *
#  *       N_VLinearSum_Serial(a,x,b,y,y);
#  *
#  *     (which stores the result of the operation a*x+b*y in y)
#  *     is legal.
#  * -----------------------------------------------------------------*/
#
# #ifndef _NVECTOR_SERIAL_H
# #define _NVECTOR_SERIAL_H
#
# #include <stdio.h>
# #include <sundials/sundials_nvector.h>
#
# #ifdef __cplusplus  /* wrapper to enable C++ usage */
# extern "C" {
# #endif
#
# /*
#  * -----------------------------------------------------------------
#  * SERIAL implementation of N_Vector
#  * -----------------------------------------------------------------
#  */
#
# struct _N_VectorContent_Serial {
#   sunindextype length;   /* vector length       */
#   booleantype own_data;  /* data ownership flag */
#   realtype *data;        /* data array          */
# };
class _N_VectorContent_Serial(Structure):
    _fields_ = [('length',   sunindextype),
                ('own_data', booleantype),
                ('data',     POINTER(realtype))]
#
# typedef struct _N_VectorContent_Serial *N_VectorContent_Serial;
N_VectorContent_Serial = POINTER(_N_VectorContent_Serial)
#
# /*
#  * -----------------------------------------------------------------
#  * Macros NV_CONTENT_S, NV_DATA_S, NV_OWN_DATA_S,
#  *        NV_LENGTH_S, and NV_Ith_S
#  * -----------------------------------------------------------------
#  */
#
# #define NV_CONTENT_S(v)  ( (N_VectorContent_Serial)(v->content) )
def NV_CONTENT_S(v):
    return cast(v.contents.content, N_VectorContent_Serial)
#
# #define NV_LENGTH_S(v)   ( NV_CONTENT_S(v)->length )
#
# #define NV_OWN_DATA_S(v) ( NV_CONTENT_S(v)->own_data )
#
# #define NV_DATA_S(v)     ( NV_CONTENT_S(v)->data )
def NV_DATA_S(v):
    return NV_CONTENT_S(v).contents.data
#
# #define NV_Ith_S(v,i)    ( NV_DATA_S(v)[i] )
#
# /*
#  * -----------------------------------------------------------------
#  * Functions exported by nvector_serial
#  * -----------------------------------------------------------------
#  */
#
# SUNDIALS_EXPORT N_Vector N_VNew_Serial(sunindextype vec_length);
N_VNew_Serial = getattr(sundials_nvecserial, 'N_VNew_Serial')
N_VNew_Serial.argtypes = [c_long]
N_VNew_Serial.restype = N_Vector
#
# SUNDIALS_EXPORT N_Vector N_VNewEmpty_Serial(sunindextype vec_length);
#
# SUNDIALS_EXPORT N_Vector N_VMake_Serial(sunindextype vec_length, realtype *v_data);
#
# SUNDIALS_EXPORT N_Vector* N_VCloneVectorArray_Serial(int count, N_Vector w);
#
# SUNDIALS_EXPORT N_Vector* N_VCloneVectorArrayEmpty_Serial(int count, N_Vector w);
#
# SUNDIALS_EXPORT void N_VDestroyVectorArray_Serial(N_Vector* vs, int count);
#
# SUNDIALS_EXPORT sunindextype N_VGetLength_Serial(N_Vector v);
#
# SUNDIALS_EXPORT void N_VPrint_Serial(N_Vector v);
#
# SUNDIALS_EXPORT void N_VPrintFile_Serial(N_Vector v, FILE *outfile);
#
# SUNDIALS_EXPORT N_Vector_ID N_VGetVectorID_Serial(N_Vector v);
# SUNDIALS_EXPORT N_Vector N_VCloneEmpty_Serial(N_Vector w);
# SUNDIALS_EXPORT N_Vector N_VClone_Serial(N_Vector w);
# SUNDIALS_EXPORT void N_VDestroy_Serial(N_Vector v);
N_VDestroy_Serial = getattr(sundials_nvecserial, 'N_VDestroy_Serial')
N_VDestroy_Serial.argtypes = [N_Vector]
N_VDestroy_Serial.restype = None
# SUNDIALS_EXPORT void N_VSpace_Serial(N_Vector v, sunindextype *lrw, sunindextype *liw);
# SUNDIALS_EXPORT realtype *N_VGetArrayPointer_Serial(N_Vector v);
# SUNDIALS_EXPORT void N_VSetArrayPointer_Serial(realtype *v_data, N_Vector v);
#
# /* standard vector operations */
# SUNDIALS_EXPORT void N_VLinearSum_Serial(realtype a, N_Vector x, realtype b, N_Vector y, N_Vector z);
# SUNDIALS_EXPORT void N_VConst_Serial(realtype c, N_Vector z);
# SUNDIALS_EXPORT void N_VProd_Serial(N_Vector x, N_Vector y, N_Vector z);
# SUNDIALS_EXPORT void N_VDiv_Serial(N_Vector x, N_Vector y, N_Vector z);
# SUNDIALS_EXPORT void N_VScale_Serial(realtype c, N_Vector x, N_Vector z);
# SUNDIALS_EXPORT void N_VAbs_Serial(N_Vector x, N_Vector z);
# SUNDIALS_EXPORT void N_VInv_Serial(N_Vector x, N_Vector z);
# SUNDIALS_EXPORT void N_VAddConst_Serial(N_Vector x, realtype b, N_Vector z);
# SUNDIALS_EXPORT realtype N_VDotProd_Serial(N_Vector x, N_Vector y);
# SUNDIALS_EXPORT realtype N_VMaxNorm_Serial(N_Vector x);
# SUNDIALS_EXPORT realtype N_VWrmsNorm_Serial(N_Vector x, N_Vector w);
# SUNDIALS_EXPORT realtype N_VWrmsNormMask_Serial(N_Vector x, N_Vector w, N_Vector id);
# SUNDIALS_EXPORT realtype N_VMin_Serial(N_Vector x);
# SUNDIALS_EXPORT realtype N_VWL2Norm_Serial(N_Vector x, N_Vector w);
# SUNDIALS_EXPORT realtype N_VL1Norm_Serial(N_Vector x);
# SUNDIALS_EXPORT void N_VCompare_Serial(realtype c, N_Vector x, N_Vector z);
# SUNDIALS_EXPORT booleantype N_VInvTest_Serial(N_Vector x, N_Vector z);
# SUNDIALS_EXPORT booleantype N_VConstrMask_Serial(N_Vector c, N_Vector x, N_Vector m);
# SUNDIALS_EXPORT realtype N_VMinQuotient_Serial(N_Vector num, N_Vector denom);
#
# /* fused vector operations */
# SUNDIALS_EXPORT int N_VLinearCombination_Serial(int nvec, realtype* c, N_Vector* V,
#                                                 N_Vector z);
# SUNDIALS_EXPORT int N_VScaleAddMulti_Serial(int nvec, realtype* a, N_Vector x,
#                                             N_Vector* Y, N_Vector* Z);
# SUNDIALS_EXPORT int N_VDotProdMulti_Serial(int nvec, N_Vector x,
#                                            N_Vector* Y, realtype* dotprods);
#
# /* vector array operations */
# SUNDIALS_EXPORT int N_VLinearSumVectorArray_Serial(int nvec,
#                                                    realtype a, N_Vector* X,
#                                                    realtype b, N_Vector* Y,
#                                                    N_Vector* Z);
# SUNDIALS_EXPORT int N_VScaleVectorArray_Serial(int nvec, realtype* c,
#                                                N_Vector* X, N_Vector* Z);
# SUNDIALS_EXPORT int N_VConstVectorArray_Serial(int nvecs, realtype c,
#                                                N_Vector* Z);
# SUNDIALS_EXPORT int N_VWrmsNormVectorArray_Serial(int nvecs, N_Vector* X,
#                                                   N_Vector* W, realtype* nrm);
# SUNDIALS_EXPORT int N_VWrmsNormMaskVectorArray_Serial(int nvecs, N_Vector* X,
#                                                       N_Vector* W, N_Vector id,
#                                                       realtype* nrm);
# SUNDIALS_EXPORT int N_VScaleAddMultiVectorArray_Serial(int nvec, int nsum,
#                                                        realtype* a,
#                                                        N_Vector* X,
#                                                        N_Vector** Y,
#                                                        N_Vector** Z);
# SUNDIALS_EXPORT int N_VLinearCombinationVectorArray_Serial(int nvec, int nsum,
#                                                            realtype* c,
#                                                            N_Vector** X,
#                                                            N_Vector* Z);
#
# /* OPTIONAL local reduction kernels (no parallel communication) */
# SUNDIALS_EXPORT realtype N_VWSqrSumLocal_Serial(N_Vector x, N_Vector w);
# SUNDIALS_EXPORT realtype N_VWSqrSumMaskLocal_Serial(N_Vector x, N_Vector w, N_Vector id);
#
# /*
#  * -----------------------------------------------------------------
#  * Enable / disable fused vector operations
#  * -----------------------------------------------------------------
#  */
#
# SUNDIALS_EXPORT int N_VEnableFusedOps_Serial(N_Vector v, booleantype tf);
#
# SUNDIALS_EXPORT int N_VEnableLinearCombination_Serial(N_Vector v, booleantype tf);
# SUNDIALS_EXPORT int N_VEnableScaleAddMulti_Serial(N_Vector v, booleantype tf);
# SUNDIALS_EXPORT int N_VEnableDotProdMulti_Serial(N_Vector v, booleantype tf);
#
# SUNDIALS_EXPORT int N_VEnableLinearSumVectorArray_Serial(N_Vector v, booleantype tf);
# SUNDIALS_EXPORT int N_VEnableScaleVectorArray_Serial(N_Vector v, booleantype tf);
# SUNDIALS_EXPORT int N_VEnableConstVectorArray_Serial(N_Vector v, booleantype tf);
# SUNDIALS_EXPORT int N_VEnableWrmsNormVectorArray_Serial(N_Vector v, booleantype tf);
# SUNDIALS_EXPORT int N_VEnableWrmsNormMaskVectorArray_Serial(N_Vector v, booleantype tf);
# SUNDIALS_EXPORT int N_VEnableScaleAddMultiVectorArray_Serial(N_Vector v, booleantype tf);
# SUNDIALS_EXPORT int N_VEnableLinearCombinationVectorArray_Serial(N_Vector v, booleantype tf);
#
# #ifdef __cplusplus
# }
# #endif
#
# #endif



""" FMI 3.0 interface """

import os
import pathlib
from ctypes import *
from . import sharedLibraryExtension, platform_tuple
from .fmi1 import _FMU, printLogMessage


fmi3Instance            = c_void_p
fmi3InstanceEnvironment = c_void_p
fmi3FMUState            = c_void_p
fmi3ValueReference      = c_uint
fmi3Float32             = c_float
fmi3Float64             = c_double
fmi3Int8                = c_int8
fmi3UInt8               = c_uint8
fmi3Int16               = c_int16
fmi3UInt16              = c_uint16
fmi3Int32               = c_int32
fmi3UInt32              = c_uint32
fmi3Int64               = c_int64
fmi3UInt64              = c_uint64
fmi3Boolean             = c_bool
fmi3Char                = c_char
fmi3String              = c_char_p
fmi3Byte                = c_char
fmi3Binary              = c_char_p
fmi3Clock               = c_bool

# values for fmi3Boolean
fmi3True  = c_bool(True)
fmi3False = c_bool(False)

# values for fmi3Clock
fmi3ClockActive   = c_bool(True)
fmi3ClockInactive = c_bool(False)

# enum fmi3Status
fmi3Status  = c_int
fmi3OK      = 0
fmi3Warning = 1
fmi3Discard = 2
fmi3Error   = 3
fmi3Fatal   = 4

# enum fmi3DependencyKind
fmi3DependencyKind = c_int
fmi3Independent    = 0
fmi3Constant       = 1
fmi3Fixed          = 2
fmi3Tunable        = 3
fmi3Discrete       = 4
fmi3Dependent      = 5

# enum fmi3IntervalQualifier
fmi3IntervalQualifier   = c_int
fmi3IntervalNotYetKnown = 0
fmi3IntervalUnchanged   = 1
fmi3IntervalChanged     = 2

# callback functions
fmi3CallbackLogMessageTYPE         = CFUNCTYPE(None, fmi3InstanceEnvironment, fmi3String, fmi3Status, fmi3String, fmi3String)
fmi3CallbackIntermediateUpdateTYPE = CFUNCTYPE(None, fmi3InstanceEnvironment, fmi3Float64, fmi3Boolean, fmi3Boolean, fmi3Boolean, fmi3Boolean, fmi3Boolean, POINTER(fmi3Boolean), POINTER(fmi3Float64))
fmi3CallbackLockPreemptionTYPE     = CFUNCTYPE(None)
fmi3CallbackUnlockPreemptionTYPE   = CFUNCTYPE(None)


def intermediateUpdate(instanceEnvironment: fmi3InstanceEnvironment,
                       intermediateUpdateTime: fmi3Float64,
                       clocksTicked: fmi3Boolean,
                       intermediateVariableSetRequested: fmi3Boolean,
                       intermediateVariableGetAllowed: fmi3Boolean,
                       intermediateStepFinished: fmi3Boolean,
                       canReturnEarly: fmi3Boolean,
                       earlyReturnRequested: POINTER(fmi3Boolean),
                       earlyReturnTime: POINTER(fmi3Float64)) -> fmi3Status:

    earlyReturnRequested.contents = fmi3False

    return fmi3OK


def printLogMessage(instanceEnvironment: fmi3InstanceEnvironment,
                    instanceName: fmi3String,
                    status: fmi3Status,
                    category: fmi3Status,
                    message: fmi3Status) -> None:
    """ Print the FMU's log messages to the command line """

    label = ['OK', 'WARNING', 'DISCARD', 'ERROR', 'FATAL', 'PENDING'][status]
    print("[%s] %s" % (label, message))


class _FMU3(_FMU):
    """ Base class for FMI 3.0 FMUs """

    def __init__(self, **kwargs):

        # build the path to the shared library
        kwargs['libraryPath'] = os.path.join(kwargs['unzipDirectory'], 'binaries', platform_tuple,
                                             kwargs['modelIdentifier'] + sharedLibraryExtension)

        super(_FMU3, self).__init__(**kwargs)

        # inquire version numbers and setting logging status
        self._fmi3Function('fmi3GetVersion', [], fmi3String)

        self._fmi3Function('fmi3SetDebugLogging', [
            (fmi3Instance,        'instance'),
            (fmi3Boolean,         'loggingOn'),
            (c_size_t,            'nCategories'),
            (POINTER(fmi3String), 'categories')
        ])

        self._fmi3Function('fmi3InstantiateModelExchange', [
            (fmi3String,                         'instanceName'),
            (fmi3String,                         'instantiationToken'),
            (fmi3String,                         'resourcePath'),
            (fmi3Boolean,                        'visible'),
            (fmi3Boolean,                        'loggingOn'),
            (fmi3InstanceEnvironment,            'instanceEnvironment'),
            (fmi3CallbackLogMessageTYPE,         'logMessage')
        ], fmi3Instance)

        self._fmi3Function('fmi3InstantiateCoSimulation', [
            (fmi3String,                         'instanceName'),
            (fmi3String,                         'instantiationToken'),
            (fmi3String,                         'resourcePath'),
            (fmi3Boolean,                        'visible'),
            (fmi3Boolean,                        'loggingOn'),
            (fmi3Boolean,                        'eventModeUsed'),
            (fmi3Boolean,                        'earlyReturnAllowed'),
            (POINTER(fmi3ValueReference),        'requiredIntermediateVariables'),
            (c_size_t,                           'nRequiredIntermediateVariables'),
            (fmi3InstanceEnvironment,            'instanceEnvironment'),
            (fmi3CallbackLogMessageTYPE,         'logMessage'),
            (fmi3CallbackIntermediateUpdateTYPE, 'intermediateUpdate')
        ], fmi3Instance)

        self._fmi3Function('fmi3InstantiateScheduledExecution', [
            (fmi3String,                         'instanceName'),
            (fmi3String,                         'instantiationToken'),
            (fmi3String,                         'resourcePath'),
            (fmi3Boolean,                        'visible'),
            (fmi3Boolean,                        'loggingOn'),
            (POINTER(fmi3ValueReference),        'requiredIntermediateVariables'),
            (c_size_t,                           'nRequiredIntermediateVariables'),
            (fmi3InstanceEnvironment,            'instanceEnvironment'),
            (fmi3CallbackLogMessageTYPE,         'logMessage'),
            (fmi3CallbackIntermediateUpdateTYPE, 'intermediateUpdate'),
            (fmi3CallbackLockPreemptionTYPE,     'lockPreemption'),
            (fmi3CallbackUnlockPreemptionTYPE,   'unlockPreemption'),
        ], fmi3Instance)

        self._fmi3Function('fmi3FreeInstance', [(fmi3Instance, 'instance')], None)

        # Enter and exit initialization mode, terminate and reset
        self._fmi3Function('fmi3EnterInitializationMode', [
            (fmi3Instance, 'instance'),
            (fmi3Boolean,  'toleranceDefined'),
            (fmi3Float64,  'tolerance'),
            (fmi3Float64,  'startTime'),
            (fmi3Boolean,  'stopTimeDefined'),
            (fmi3Float64,  'stopTime')
        ])

        self._fmi3Function('fmi3ExitInitializationMode', [(fmi3Instance, 'instance')])

        self._fmi3Function('fmi3EnterEventMode', [
            (fmi3Instance,       'instance'),
            (fmi3Boolean,        'stepEvent'),
            (fmi3Boolean,        'stateEvent'),
            (POINTER(fmi3Int32), 'rootsFound'),
            (c_size_t,           'nEventIndicators'),
            (fmi3Boolean,        'timeEvent'),
        ])

        self._fmi3Function('fmi3Terminate', [(fmi3Instance, 'instance')])

        self._fmi3Function('fmi3Reset', [(fmi3Instance, 'instance')])

        # Getting and setting variable values
        types = [
            ('Float32', fmi3Float32),
            ('Float64', fmi3Float64),
            ('Int8',    fmi3Int8),
            ('UInt8',   fmi3UInt8),
            ('Int16',   fmi3Int16),
            ('UInt16',  fmi3UInt16),
            ('Int32',   fmi3Int32),
            ('UInt32',  fmi3UInt32),
            ('Int64',   fmi3Int64),
            ('UInt64',  fmi3UInt64),
            ('Boolean', fmi3Boolean),
            ('String',  fmi3String),
            ('Clock',   fmi3Clock),
        ]

        for name, _type in types:

            params = [
                (fmi3Instance,                'instance'),
                (POINTER(fmi3ValueReference), 'vr'),
                (c_size_t,                    'nvr'),
                (POINTER(_type),              'value'),
                (c_size_t,                    'nValues')
            ]

            self._fmi3Function('fmi3Get' + name, params)
            self._fmi3Function('fmi3Set' + name, params)

        self._fmi3Function('fmi3GetBinary', [
            (fmi3Instance,                'instance'),
            (POINTER(fmi3ValueReference), 'vr'),
            (c_size_t,                    'nvr'),
            (POINTER(c_size_t),           'size'),
            (POINTER(fmi3Binary),         'value'),
            (c_size_t,                    'nValues')
        ])

        self._fmi3Function('fmi3SetBinary', [
            (fmi3Instance,                'instance'),
            (POINTER(fmi3ValueReference), 'vr'),
            (c_size_t,                    'nvr'),
            (POINTER(c_size_t),           'size'),
            (POINTER(fmi3Binary),         'value'),
            (c_size_t,                    'nValues')
        ])

        # Getting Variable Dependency Information
        self._fmi3Function('fmi3GetNumberOfVariableDependencies', [
            (fmi3Instance,       'instance'),
            (fmi3ValueReference, 'valueReference'),
            (POINTER(c_size_t),  'nDependencies')
        ])

        self._fmi3Function('fmi3GetVariableDependencies', [
            (fmi3Instance,                'instance'),
            (fmi3ValueReference,          'dependent'),
            (POINTER(c_size_t),           'elementIndicesOfDependent'),
            (POINTER(fmi3ValueReference), 'independents'),
            (POINTER(c_size_t),           'elementIndicesOfIndependents'),
            (POINTER(fmi3DependencyKind), 'dependencyKinds'),
            (c_size_t,                    'nDependencies')
        ])

        # Getting and setting the internal FMU state
        self._fmi3Function('fmi3GetFMUState', [(fmi3Instance, 'instance'), (POINTER(fmi3FMUState), 'FMUState')])

        self._fmi3Function('fmi3SetFMUState', [(fmi3Instance, 'instance'), (fmi3FMUState, 'FMUState')])

        self._fmi3Function('fmi3FreeFMUState', [(fmi3Instance, 'instance'), (POINTER(fmi3FMUState), 'FMUState')])

        self._fmi3Function('fmi3SerializedFMUStateSize', [
            (fmi3Instance,      'instance'),
            (fmi3FMUState,      'FMUState'),
            (POINTER(c_size_t), 'size')
        ])

        self._fmi3Function('fmi3SerializeFMUState', [
            (fmi3Instance,      'instance'),
            (fmi3FMUState,      'FMUState'),
            (POINTER(fmi3Byte), 'serializedState'),
            (c_size_t,          'size')
        ])

        self._fmi3Function('fmi3DeSerializeFMUState', [
            (fmi3Instance,          'instance'),
            (POINTER(fmi3Byte),     'serializedState'),
            (c_size_t,              'size'),
            (POINTER(fmi3FMUState), 'FMUState'),
        ])

        # Getting partial derivatives
        self._fmi3Function('fmi3GetDirectionalDerivative', [
            (fmi3Instance,                'instance'),
            (POINTER(fmi3ValueReference), 'unknowns'),
            (c_size_t,                    'nUnknowns'),
            (POINTER(fmi3ValueReference), 'knowns'),
            (c_size_t,                    'nKnowns'),
            (POINTER(fmi3Float64),        'seed'),
            (c_size_t,                    'nSeed'),
            (POINTER(fmi3Float64),        'sensitivity'),
            (c_size_t,                    'nSensitivity')
        ])

        self._fmi3Function('fmi3GetAdjointDerivative', [
            (fmi3Instance,                'instance'),
            (POINTER(fmi3ValueReference), 'unknowns'),
            (c_size_t,                    'nUnknowns'),
            (POINTER(fmi3ValueReference), 'knowns'),
            (c_size_t,                    'nKnowns'),
            (POINTER(fmi3Float64),        'seed'),
            (c_size_t,                    'nSeed'),
            (POINTER(fmi3Float64),        'sensitivity'),
            (c_size_t,                    'nSensitivity')
        ])

        # Entering and exiting the Configuration or Reconfiguration Mode
        self._fmi3Function('fmi3EnterConfigurationMode', [(fmi3Instance, 'instance')])

        self._fmi3Function('fmi3ExitConfigurationMode', [(fmi3Instance, 'instance')])

        # Clock related functions
        self._fmi3Function('fmi3GetIntervalDecimal', [
            (fmi3Instance,                   'instance'),
            (POINTER(fmi3ValueReference),    'valueReferences'),
            (c_size_t,                       'nValueReferences'),
            (POINTER(fmi3Float64),           'intervals'),
            (POINTER(fmi3IntervalQualifier), 'qualifiers'),
            (c_size_t,                       'nIntervals'),
        ])

        self._fmi3Function('fmi3GetIntervalFraction', [
            (fmi3Instance,                   'instance'),
            (POINTER(fmi3ValueReference),    'valueReferences'),
            (c_size_t,                       'nValueReferences'),
            (POINTER(fmi3UInt64),            'intervalCounters'),
            (POINTER(fmi3UInt64),            'resolutions'),
            (POINTER(fmi3IntervalQualifier), 'qualifiers'),
            (c_size_t,                       'nIntervals')
        ])

        self._fmi3Function('fmi3GetShiftDecimal', [
            (fmi3Instance,                   'instance'),
            (POINTER(fmi3ValueReference),    'valueReferences'),
            (c_size_t,                       'nValueReferences'),
            (POINTER(fmi3Float64),           'shifts'),
            (c_size_t,                       'nShifts'),
        ])

        self._fmi3Function('fmi3GetShiftFraction', [
            (fmi3Instance,                   'instance'),
            (POINTER(fmi3ValueReference),    'valueReferences'),
            (c_size_t,                       'nValueReferences'),
            (POINTER(fmi3UInt64),            'shiftCounters'),
            (POINTER(fmi3UInt64),            'resolutions'),
            (c_size_t,                       'nShifts')
        ])

        self._fmi3Function('fmi3SetIntervalDecimal', [
            (fmi3Instance,                'instance'),
            (POINTER(fmi3ValueReference), 'valueReferences'),
            (c_size_t,                    'nValueReferences'),
            (POINTER(fmi3Float64),        'intervals'),
            (c_size_t,                    'nIntervals')
        ])

        self._fmi3Function('fmi3SetIntervalFraction', [
            (fmi3Instance,                'instance'),
            (POINTER(fmi3ValueReference), 'valueReferences'),
            (c_size_t,                    'nValueReferences'),
            (POINTER(fmi3UInt64),         'intervalCounters'),
            (POINTER(fmi3UInt64),         'resolutions'),
            (c_size_t,                    'nIntervals')
        ])

        self._fmi3Function('fmi3EvaluateDiscreteStates', [
            (fmi3Instance, 'instance')
        ])

        self._fmi3Function('fmi3UpdateDiscreteStates', [
            (fmi3Instance,         'instance'),
            (POINTER(fmi3Boolean), 'discreteStatesNeedUpdate'),
            (POINTER(fmi3Boolean), 'terminateSimulation'),
            (POINTER(fmi3Boolean), 'nominalsOfContinuousStatesChanged'),
            (POINTER(fmi3Boolean), 'valuesOfContinuousStatesChanged'),
            (POINTER(fmi3Boolean), 'nextEventTimeDefined'),
            (POINTER(fmi3Float64), 'nextEventTime')
        ])

        # Functions for Model Exchange

        self._fmi3Function('fmi3EnterContinuousTimeMode', [(fmi3Instance, 'instance')])

        self._fmi3Function('fmi3CompletedIntegratorStep', [
            (fmi3Instance,         'instance'),
            (fmi3Boolean,          'noSetFMUStatePriorToCurrentPoint'),
            (POINTER(fmi3Boolean), 'enterEventMode'),
            (POINTER(fmi3Boolean), 'terminateSimulation')
        ])

        #  Providing independent variables and re-initialization of caching

        self._fmi3Function('fmi3SetTime', [
            (fmi3Instance, 'instance'),
            (fmi3Float64,  'time')
        ])

        self._fmi3Function('fmi3SetContinuousStates', [
            (fmi3Instance,         'instance'),
            (POINTER(fmi3Float64), 'continuousStates'),
            (c_size_t,             'nContinuousStates')
        ])

        # Evaluation of the model equations

        self._fmi3Function('fmi3GetContinuousStateDerivatives', [
            (fmi3Instance,         'instance'),
            (POINTER(fmi3Float64), 'derivatives'),
            (c_size_t,             'nContinuousStates')
        ])

        self._fmi3Function('fmi3GetEventIndicators', [
            (fmi3Instance,         'instance'),
            (POINTER(fmi3Float64), 'eventIndicators'),
            (c_size_t,             'ni')
        ])

        self._fmi3Function('fmi3GetContinuousStates', [
            (fmi3Instance,         'instance'),
            (POINTER(fmi3Float64), 'continuousStates'),
            (c_size_t,             'nContinuousStates')
        ])

        self._fmi3Function('fmi3GetNominalsOfContinuousStates', [
            (fmi3Instance,         'instance'),
            (POINTER(fmi3Float64), 'nominals'),
            (c_size_t,             'nContinuousStates')
        ])

        self._fmi3Function('fmi3GetNumberOfEventIndicators', [
            (fmi3Instance,      'instance'),
            (POINTER(c_size_t), 'nEventIndicators')
        ])

        self._fmi3Function('fmi3GetNumberOfContinuousStates', [
            (fmi3Instance,      'instance'),
            (POINTER(c_size_t), 'nContinuousStates')
        ])

        # Functions for Co-Simulation

        # Simulating the FMU

        self._fmi3Function('fmi3EnterStepMode', [(fmi3Instance, 'instance')])

        self._fmi3Function('fmi3GetOutputDerivatives', [
            (fmi3Instance, 'instance'),
            (POINTER(fmi3ValueReference), 'valueReferences'),
            (c_size_t, 'nValueReferences'),
            (POINTER(fmi3Int32), 'orders'),
            (POINTER(fmi3Float64), 'values'),
            (c_size_t, 'nValues'),
        ])

        self._fmi3Function('fmi3DoStep', [
            (fmi3Instance, 'instance'),
            (fmi3Float64, 'currentCommunicationPoint'),
            (fmi3Float64, 'communicationStepSize'),
            (fmi3Boolean, 'noSetFMUStatePriorToCurrentPoint'),
            (POINTER(fmi3Boolean), 'eventEncountered'),
            (POINTER(fmi3Boolean), 'terminateSimulation'),
            (POINTER(fmi3Boolean), 'earlyReturn'),
            (POINTER(fmi3Float64), 'lastSuccessfulTime')
        ])

        self._fmi3Function('fmi3ActivateModelPartition', [
            (fmi3Instance, 'instance'),
            (fmi3ValueReference, 'clockReference'),
            (c_size_t, 'clockElementIndex'),
            (fmi3Float64, 'activationTime')
        ])

    def _fmi3Function(self, fname, params, restype=fmi3Status):
        """ Add an FMI 3.0 function to this instance and add a wrapper that allows
        logging and checks the return code if the return type is fmi3Status

        Parameters:
            fname     the name of the function
            params    parameters as (type, name) tuples
            restype   return type
        """

        if not hasattr(self.dll, fname):

            def raise_exception(*args):
                raise Exception("Function %s is missing in shared library." % fname)

            setattr(self, fname, raise_exception)

            return

        if len(params) > 0:
            argtypes, argnames = zip(*params)
        else:
            argtypes = argnames = []

        # get the exported function form the shared library
        f = getattr(self.dll, fname)
        f.argtypes = argtypes
        f.restype = restype

        def w(*args):
            """ Wrapper function for the FMI call """

            # call the FMI function
            res = f(*args)

            if self.fmiCallLogger is not None:
                # log the call
                self._log_fmi_args(fname, argnames, argtypes, args, restype, res)

            if restype == fmi3Status:  # status code
                # check the status code
                if res > fmi3Warning:
                    raise Exception("%s failed with status %d." % (fname, res))

            return res

        setattr(self, fname, w)

    # Inquire version numbers of header files and setting logging status

    def getVersion(self):
        version = self.fmi3GetVersion()
        return version.decode('utf-8')

    def setDebugLogging(self, loggingOn, categories):
        categories_ = (fmi3String * len(categories))()
        categories_[:] = [c.encode('utf-8') for c in categories]
        self.fmi3SetDebugLogging(self.component, fmi3True if loggingOn else fmi3False, len(categories), categories_)

    # Creation and destruction of FMU instances and setting debug status

    def freeInstance(self):
        self.fmi3FreeInstance(self.component)
        self.freeLibrary()

    # Enter and exit initialization mode, terminate and reset

    def enterInitializationMode(self, tolerance=None, startTime=0.0, stopTime=None):

        toleranceDefined = fmi3True if tolerance is not None else fmi3False

        if tolerance is None:
            tolerance = 0.0

        stopTimeDefined = fmi3True if stopTime is not None else fmi3False

        if stopTime is None:
            stopTime = 0.0

        return self.fmi3EnterInitializationMode(self.component, toleranceDefined, tolerance, startTime, stopTimeDefined, stopTime)

    def exitInitializationMode(self):
        return self.fmi3ExitInitializationMode(self.component)

    # Clock related functions

    def getIntervalDecimal(self, valueReferences, intervals, qualifiers):
        self.fmi3GetIntervalDecimal(self.component, valueReferences, len(valueReferences), intervals, qualifiers, len(intervals))

    def getIntervalFraction(self, valueReferences, intervalCounters, resolutions, qualifiers):
        self.fmi3GetIntervalFraction(self.component, valueReferences, len(valueReferences), intervalCounters, resolutions, qualifiers, len(intervalCounters))

    def getShiftDecimal(self, valueReferences, shifts):
        self.fmi3GetShiftDecimal(self.component, valueReferences, len(valueReferences), shifts, len(shifts))

    def getShiftFraction(self, valueReferences, shiftCounters, resolutions):
        self.fmi3GetShiftFraction(self.component, valueReferences, len(valueReferences), shiftCounters, resolutions, len(shiftCounters))

    def setIntervalDecimal(self, valueReferences, intervals):
        self.fmi3SetIntervalDecimal(self.component, valueReferences, len(valueReferences), intervals, len(intervals))

    def setIntervalFraction(self, valueReferences, intervalCounters, resolutions):
        self.fmi3SetIntervalFraction(self.component, valueReferences, len(valueReferences), intervalCounters, resolutions, len(intervalCounters))

    def enterEventMode(self, stepEvent=False, stateEvent=False, rootsFound=[], timeEvent=False):

        rootsFound = (fmi3Int32 * len(rootsFound))(*rootsFound)

        return self.fmi3EnterEventMode(
            self.component,
            fmi3True if stepEvent else fmi3False,
            fmi3True if stateEvent else fmi3False,
            rootsFound,
            len(rootsFound),
            fmi3True if timeEvent else fmi3False,
        )

    def updateDiscreteStates(self):

        discreteStatesNeedUpdate          = fmi3Boolean()
        terminateSimulation               = fmi3Boolean()
        nominalsOfContinuousStatesChanged = fmi3Boolean()
        valuesOfContinuousStatesChanged   = fmi3Boolean()
        nextEventTimeDefined              = fmi3Boolean()
        nextEventTime                     = fmi3Float64()

        self.fmi3UpdateDiscreteStates(self.component,
                                      byref(discreteStatesNeedUpdate),
                                      byref(terminateSimulation),
                                      byref(nominalsOfContinuousStatesChanged),
                                      byref(valuesOfContinuousStatesChanged),
                                      byref(nextEventTimeDefined),
                                      byref(nextEventTime))

        return (discreteStatesNeedUpdate.value          != fmi3False,
                terminateSimulation.value               != fmi3False,
                nominalsOfContinuousStatesChanged.value != fmi3False,
                valuesOfContinuousStatesChanged.value   != fmi3False,
                nextEventTimeDefined.value              != fmi3False,
                nextEventTime.value)

    def terminate(self):
        return self.fmi3Terminate(self.component)

    def reset(self):
        return self.fmi3Reset(self.component)

    # Getting and setting variable values

    def getFloat32(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3Float32 * nValues)()
        self.fmi3GetFloat32(self.component, vr, len(vr), values, nValues)
        return list(values)

    def getFloat64(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3Float64 * nValues)()
        self.fmi3GetFloat64(self.component, vr, len(vr), values, nValues)
        return list(values)

    def getInt8(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3Int8 * nValues)()
        self.fmi3GetInt8(self.component, vr, len(vr), value, nValues)
        return list(value)

    def getUInt8(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3UInt8 * nValues)()
        self.fmi3GetUInt8(self.component, vr, len(vr), value, nValues)
        return list(value)

    def getInt16(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3Int16 * nValues)()
        self.fmi3GetInt16(self.component, vr, len(vr), value, nValues)
        return list(value)

    def getUInt16(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3UInt16 * nValues)()
        self.fmi3GetUInt16(self.component, vr, len(vr), value, nValues)
        return list(value)

    def getInt32(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3Int32 * nValues)()
        self.fmi3GetInt32(self.component, vr, len(vr), value, nValues)
        return list(value)

    def getUInt32(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3UInt32 * nValues)()
        self.fmi3GetUInt32(self.component, vr, len(vr), value, nValues)
        return list(value)

    def getInt64(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3Int64 * nValues)()
        self.fmi3GetInt64(self.component, vr, len(vr), value, nValues)
        return list(value)

    def getUInt64(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3UInt64 * nValues)()
        self.fmi3GetUInt64(self.component, vr, len(vr), value, nValues)
        return list(value)

    def getBoolean(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3Boolean * nValues)()
        self.fmi3GetBoolean(self.component, vr, len(vr), value, nValues)
        return list(value)

    def getString(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3String * nValues)()
        self.fmi3GetString(self.component, vr, len(vr), value, nValues)
        return list(map(lambda b: b.decode('utf-8'), value))

    def getBinary(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3Binary * nValues)()
        size = (c_size_t * len(vr))()
        self.fmi3GetBinary(self.component, vr, len(vr), size, value, nValues)
        return list(value)

    def getClock(self, vr, nValues=None):
        if nValues is None:
            nValues = len(vr)
        vr = (fmi3ValueReference * len(vr))(*vr)
        value = (fmi3Clock * nValues)()
        self.fmi3GetClock(self.component, vr, len(vr), value, nValues)
        return list(value)

    def setFloat32(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3Float32 * len(values))(*values)
        self.fmi3SetFloat32(self.component, vr, len(vr), values, len(values))

    def setFloat64(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3Float64 * len(values))(*values)
        self.fmi3SetFloat64(self.component, vr, len(vr), values, len(values))

    def setInt8(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3Int8 * len(values))(*values)
        self.fmi3SetInt8(self.component, vr, len(vr), values, len(values))

    def setUInt8(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3UInt8 * len(values))(*values)
        self.fmi3SetUInt8(self.component, vr, len(vr), values, len(values))

    def setInt16(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3Int16 * len(values))(*values)
        self.fmi3SetInt16(self.component, vr, len(vr), values, len(values))

    def setUInt16(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3UInt16 * len(values))(*values)
        self.fmi3SetUInt16(self.component, vr, len(vr), values, len(values))

    def setInt32(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3Int32 * len(values))(*values)
        self.fmi3SetInt32(self.component, vr, len(vr), values, len(values))

    def setUInt32(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3UInt32 * len(values))(*values)
        self.fmi3SetUInt32(self.component, vr, len(vr), values, len(values))

    def setInt64(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3Int64 * len(values))(*values)
        self.fmi3SetInt64(self.component, vr, len(vr), values, len(values))

    def setUInt64(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3UInt64 * len(values))(*values)
        self.fmi3SetUInt64(self.component, vr, len(vr), values, len(values))

    def setBoolean(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3Boolean * len(values))(*values)
        self.fmi3SetBoolean(self.component, vr, len(vr), values, len(values))

    def setString(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = list(map(lambda s: s.encode('utf-8') if s is not None else s, values))
        values = (fmi3String * len(values))(*values)
        self.fmi3SetString(self.component, vr, len(vr), values, len(values))

    def setBinary(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values_ = (fmi3Binary * len(values))(*values)
        size = (c_size_t * len(vr))(*[len(v) for v in values])
        self.fmi3SetBinary(self.component, vr, len(vr), size, values_, len(values))

    def setClock(self, vr, values):
        vr = (fmi3ValueReference * len(vr))(*vr)
        values = (fmi3Clock * len(values))(*values)
        self.fmi3SetClock(self.component, vr, len(vr), values, len(values))

    # Getting and setting the internal FMU state

    def getFMUState(self):
        state = fmi3FMUState()
        self.fmi3GetFMUState(self.component, byref(state))
        return state

    def setFMUState(self, state):
        self.fmi3SetFMUState(self.component, state)

    def freeFMUState(self, state):
        self.fmi3FreeFMUState(self.component, byref(state))

    def serializeFMUState(self, state):
        """ Serialize an FMU state

        Parameters:
            state   the FMU state

        Returns:
            the serialized state as a byte string
        """

        size = c_size_t()
        self.fmi3SerializedFMUStateSize(self.component, state, byref(size))
        serializedState = create_string_buffer(size.value)
        self.fmi3SerializeFMUState(self.component, state, serializedState, size)
        return serializedState.raw

    def deSerializeFMUState(self, serializedState, state):
        """ De-serialize an FMU state

        Parameters:
            serializedState   the serialized state as a byte string
            state             the FMU state
        """

        buffer = create_string_buffer(serializedState)
        self.fmi3DeSerializeFMUState(self.component, buffer, len(buffer), byref(state))

    # Getting partial derivatives

    def getDirectionalDerivative(self, unknowns, knowns, seed, sensitivity):
        """ Get the directional derivatives

        Parameters:
            unknowns     list of value references of the unknowns
            knowns       list of value references of the knowns
            seed         list of delta values (one per known)

        Returns:
            sensitivity  list of the partial derivatives (one per unknown)
        """

        unknowns    = (fmi3ValueReference * len(unknowns))(*unknowns)
        knowns      = (fmi3ValueReference * len(knowns))(*knowns)
        seed        = (fmi3Float64 * len(seed))(*seed)
        sensitivity = (fmi3Float64 * len(sensitivity))()

        self.fmi3GetDirectionalDerivative(self.component, unknowns, len(unknowns), knowns, len(knowns), seed, sensitivity)

        return list(sensitivity)

    def getAdjointDerivative(self, unknowns, knowns, seed, sensitivity):
        """ Get adjoint derivatives

        Parameters:
            unknowns     list of value references of the unknowns
            knowns       list of value references of the knowns
            seed         list of delta values (one per known)

        Returns:
            sensitivity  list of the partial derivatives (one per unknown)
        """

        unknowns    = (fmi3ValueReference * len(unknowns))(*unknowns)
        knowns      = (fmi3ValueReference * len(knowns))(*knowns)
        seed        = (fmi3Float64 * len(seed))(*seed)
        sensitivity = (fmi3Float64 * len(sensitivity))()

        self.fmi3GetAdjointDerivative(self.component, unknowns, len(unknowns), knowns, len(knowns), seed, sensitivity)

        return list(sensitivity)


class FMU3Model(_FMU3):
    """ An FMI 3.0 Model Exchange FMU """

    def __init__(self, **kwargs):
        super(FMU3Model, self).__init__(**kwargs)

    def instantiate(self, visible=False, loggingOn=False):

        resourcePath = os.path.join(self.unzipDirectory, 'resources')

        # save callbacks from GC
        self.printLogMessage = fmi3CallbackLogMessageTYPE(printLogMessage)

        self.component = self.fmi3InstantiateModelExchange(
            self.instanceName.encode('utf-8'),
            self.guid.encode('utf-8'),
            resourcePath.encode('utf-8'),
            fmi3True if visible else fmi3False,
            fmi3True if loggingOn else fmi3False,
            fmi3InstanceEnvironment(),
            self.printLogMessage)

        if not self.component:
            raise Exception("Failed to instantiate FMU")

    # Enter and exit the different modes

    def enterContinuousTimeMode(self):
        return self.fmi3EnterContinuousTimeMode(self.component)

    def completedIntegratorStep(self, noSetFMUStatePriorToCurrentPoint=fmi3True):
        enterEventMode = fmi3Boolean()
        terminateSimulation = fmi3Boolean()
        self.fmi3CompletedIntegratorStep(self.component, noSetFMUStatePriorToCurrentPoint, byref(enterEventMode), byref(terminateSimulation))
        return enterEventMode.value, terminateSimulation.value

    # Providing independent variables and re-initialization of caching

    def setTime(self, time):
        return self.fmi3SetTime(self.component, time)

    def setContinuousStates(self, continuousStates, nContinuousStates):
        return self.fmi3SetContinuousStates(self.component, continuousStates, nContinuousStates)

    # Evaluation of the model equations

    def getContinuousStateDerivatives(self, derivatives, nContinuousStates):
        return self.fmi3GetContinuousStateDerivatives(self.component, derivatives, nContinuousStates)

    def getEventIndicators(self, eventIndicators, nEventIndicators):
        return self.fmi3GetEventIndicators(self.component, eventIndicators, nEventIndicators)

    def getContinuousStates(self, continuousStates, nContinuousStates):
        return self.fmi3GetContinuousStates(self.component, continuousStates, nContinuousStates)

    def getNominalsOfContinuousState(self, nominals, nContinuousStates):
        return self.fmi3GetNominalsOfContinuousState(self.component, nominals, nContinuousStates)


class FMU3Slave(_FMU3):
    """ An FMI 3.0 Co-Simulation FMU """

    def __init__(self, instanceName=None, **kwargs):

        kwargs['instanceName'] = instanceName

        super(FMU3Slave, self).__init__(**kwargs)

    def instantiate(self, visible=False, loggingOn=False, eventModeUsed=False, earlyReturnAllowed=False):

        # save callbacks from GC
        self.printLogMessage = fmi3CallbackLogMessageTYPE(printLogMessage)
        self.intermediateUpdate = fmi3CallbackIntermediateUpdateTYPE(intermediateUpdate)

        resourcePath = os.path.join(self.unzipDirectory, 'resources')

        self.component = self.fmi3InstantiateCoSimulation(
            self.instanceName.encode('utf-8'),
            self.guid.encode('utf-8'),
            resourcePath.encode('utf-8'),
            fmi3True if visible else fmi3False,
            fmi3True if loggingOn else fmi3False,
            fmi3True if eventModeUsed else fmi3False,
            fmi3True if earlyReturnAllowed else fmi3False,
            None, 0,
            fmi3InstanceEnvironment(),
            self.printLogMessage,
            self.intermediateUpdate)

        if not self.component:
            raise Exception("Failed to instantiate FMU")

    # Simulating the FMU

    def enterStepMode(self):
        return self.fmi3EnterStepMode(self.component)

    def setInputDerivatives(self, vr, order, value):
        vr = (fmi3ValueReference * len(vr))(*vr)
        order = (fmi3Int32 * len(vr))(*order)
        value = (fmi3Float64 * len(vr))(*value)
        self.fmi3SetInputDerivatives(self.component, vr, len(vr), order, value)

    def getOutputDerivatives(self, vr, order):
        vr = (fmi3ValueReference * len(vr))(*vr)
        order = (fmi3Int32 * len(vr))(*order)
        value = (fmi3Float64 * len(vr))()
        self.fmi3GetOutputDerivatives(self.component, vr, len(vr), order, value)
        return list(value)

    def doStep(self, currentCommunicationPoint, communicationStepSize, noSetFMUStatePriorToCurrentPoint=fmi3True):
        eventEncountered = fmi3Boolean()
        terminateSimulation = fmi3Boolean()
        earlyReturn = fmi3Boolean()
        lastSuccessfulTime = fmi3Float64()
        status = self.fmi3DoStep(self.component, currentCommunicationPoint, communicationStepSize, noSetFMUStatePriorToCurrentPoint, byref(eventEncountered), byref(terminateSimulation), byref(earlyReturn), byref(lastSuccessfulTime))
        return status, eventEncountered.value != fmi3False, terminateSimulation.value != fmi3False, earlyReturn.value != fmi3False, lastSuccessfulTime.value


class FMU3ScheduledExecution(_FMU3):
    """ An FMI 3.0 Scheduled Execution FMU """

    def __init__(self, instanceName=None, **kwargs):
        kwargs['instanceName'] = instanceName

        super(FMU3ScheduledExecution, self).__init__(**kwargs)

        self._fmi3Function('fmi3ActivateModelPartition', [
            (fmi3Instance, 'instance'),
            (fmi3ValueReference, 'clockReference'),
            (c_size_t, 'clockElementIndex'),
            (fmi3Float64, 'activationTime'),
        ])

    def instantiate(self, visible=False, loggingOn=False):

        resourcePath = os.path.join(self.unzipDirectory, 'resources')

        def noop():
            pass

        # save callbacks from GC
        self.printLogMessage = fmi3CallbackLogMessageTYPE(printLogMessage)
        self.intermediateUpdate = fmi3CallbackIntermediateUpdateTYPE(intermediateUpdate)
        self.lockPreemption = fmi3CallbackLockPreemptionTYPE(noop)
        self.unlockPreemption = fmi3CallbackUnlockPreemptionTYPE(noop)

        self.component = self.fmi3InstantiateScheduledExecution(
            self.instanceName.encode('utf-8'),
            self.guid.encode('utf-8'),
            resourcePath.encode('utf-8'),
            fmi3True if visible else fmi3False,
            fmi3True if loggingOn else fmi3False,
            None, 0,
            fmi3InstanceEnvironment(),
            self.printLogMessage,
            self.intermediateUpdate,
            self.lockPreemption,
            self.unlockPreemption
        )

        if not self.component:
            raise Exception("Failed to instantiate FMU")

    def activateModelPartition(self, clockReference, clockElementIndex, activationTime):
        self.fmi3ActivateModelPartition(self.component, clockReference, clockElementIndex, activationTime)



from __future__ import print_function

import sys
import os
import numpy as np
import shutil

from fmpy import read_model_description
from fmpy.cross_check import validate_result
from fmpy.util import read_ref_opt_file


def read_csv(filename, variable_names=None):
    """ Read a CSV file that conforms to the FMI cross-check rules

    Parameters:
        filename         name of the CSV file to read
        variable_names   list of legal variable names

    Returns:
        traj             the trajectoies read from the CSV file
    """

    # pass an empty string as deletechars to preserve special characters
    traj = np.genfromtxt(filename, delimiter=',', names=True, deletechars='')

    # get the time
    time = traj[traj.dtype.names[0]]

    # check if the time is monotonically increasing
    if traj.size > 1 and np.any(np.diff(time) < 0):
        raise Exception("Values in first column (time) are not monotonically increasing")

    # check if all variables exist in the FMU
    if variable_names is not None:
        for name in traj.dtype.names[1:]:
            if name not in variable_names:
                raise Exception("Variable '%s' does not exist in the FMU" % name)

    return traj


def validate_test_fmu(model_dir):
    """ Validate an exported FMU

    Parameters:
        model_dir  path to the directory that contains the exported FMU

    Returns:
        a list of problems
    """

    problems = []

    # check file sizes
    for root, dirs, files in os.walk(model_dir):
        for file in files:
            filename = os.path.join(root, file)
            filesize = os.path.getsize(filename)
            maxsize = 10e6 if file.endswith('.fmu') else 1e6
            if filesize > maxsize:
                problems.append("%s is larger than %g MB (%g MB)" % (filename, maxsize * 1e-6, filesize * 1e-6))

    if 'notCompliantWithLatestRules' in files:
        return problems  # stop here

    path, model_name = os.path.split(model_dir)
    path, _ = os.path.split(path)
    path, _ = os.path.split(path)
    path, platform = os.path.split(path)
    path, fmi_type = os.path.split(path)
    _, fmi_version = os.path.split(path)

    fmu_filename = os.path.join(model_dir, model_name + '.fmu')

    # validate the modelDescription.xml
    try:
        model_description = read_model_description(fmu_filename, validate=True)
    except Exception as e:
        problems.append("Error in %s. %s" % (fmu_filename, e))
        return problems  # stop here

    # check FMI version
    if model_description.fmiVersion != fmi_version:
        problems.append("%s is not an FMI %s FMU" % (fmu_filename, fmi_version))
        return problems  # stop here

    # check FMI type
    if fmi_type == 'cs' and model_description.coSimulation is None:
        problems.append("%s does not support co-simulation" % fmu_filename)
        return problems  # stop here
    elif fmi_type == 'me' and model_description.modelExchange is None:
        problems.append("%s does not support model-exchange" % fmu_filename)
        return problems  # stop here

    # collect the variable names
    variable_names = [v.name for v in model_description.modelVariables]

    # check the reference options file
    try:
        ref_opts_filename = os.path.join(model_dir, model_name + '_ref.opt')
        read_ref_opt_file(ref_opts_filename)
    except Exception as e:
        problems.append("Error in %s. %s" % (ref_opts_filename, e))

    # check the CSVs
    for suffix, required in [('_in.csv', False), ('_ref.csv', True)]:

        csv_filename = os.path.join(model_dir, model_name + suffix)

        if not required and not os.path.isfile(csv_filename):
            continue

        try:
            read_csv(csv_filename, variable_names=variable_names)
        except Exception as e:
            problems.append("Error in %s. %s" % (csv_filename, e))

    return problems


def validate_cross_check_result(result_dir):
    """ Validate a cross-check result

    Parameters:
        result_dir  path to the directory that contains the results

    Returns:
        a list of problems
    """

    problems = []

    t = segments(result_dir)

    fmi_version, fmi_type, platform, importing_tool_name, importing_tool_version, exporting_tool_name, exporting_tool_version, model_name = t[-8:]

    repo_dir = os.path.join(*t[:-9])

    fmu_dir = os.path.join(repo_dir, 'fmus', fmi_version, fmi_type, platform, exporting_tool_name, exporting_tool_version, model_name)

    ref_filename = os.path.join(fmu_dir, model_name + '_ref.csv')
    opt_filename = os.path.join(fmu_dir, model_name + '_ref.opt')

    # check file sizes
    for root, dirs, files in os.walk(result_dir):
        for file in files:
            filename = os.path.join(root, file)
            filesize = os.path.getsize(filename)
            if filesize > 1e6:
                problems.append("%s is larger than 1 MB (%.1f MB)" % (filename, filesize * 1e-6))

    _, model_name = os.path.split(result_dir)

    not_compliant_file = os.path.join(result_dir, 'notCompliantWithLatestRules')
    passed_file = os.path.join(result_dir, 'passed')

    if os.path.isfile(not_compliant_file) or not os.path.isfile(passed_file):
        return problems  # stop here

    # check the output file
    res_filename = os.path.join(result_dir, model_name + '_out.csv')

    try:
        result = read_csv(res_filename)
    except Exception as e:
        problems.append("Error in %s. %s" % (res_filename, e))
        return problems  # stop here

    try:
        reference = read_csv(ref_filename)
        opt = read_ref_opt_file(opt_filename)
    except Exception as e:
        problems.append("Error in %s. %s" % (res_filename, e))
        return problems  # stop here

    problem = validate_result(result=result, reference=reference, t_start=opt['StartTime'], t_stop=opt['StopTime'])

    if problem is not None:
        problems.append("Error in %s. %s" % (res_filename, problem))

    return problems


def segments(path):
    """ Split a path into segments """

    s = []

    head, tail = os.path.split(path)

    while tail:
        s.insert(0, tail)
        head, tail = os.path.split(head)

    s.insert(0, head)

    return s


def validate_repo(vendor_dir, clean_up=False):

    problems = []

    s = segments(vendor_dir)

    result_count = 0

    # validate the cross-check results
    for subdir, dirs, files in os.walk(os.path.join(vendor_dir, 'results')):

        t = segments(subdir)

        if len(t) - len(s) != 9:
            continue

        result_count += 1

        fmi_version, fmi_type, platform, importing_tool_name, importing_tool_version, exporting_tool_name, exporting_tool_version, model_name = t[-8:]

        if fmi_version not in ['1.0', '2.0']:
            continue

        if fmi_type not in ['cs', 'me']:
            continue

        if platform not in ['c-code', 'darwin64', 'linux32', 'linux64', 'win32', 'win64']:
            continue

        new_problems = validate_cross_check_result(subdir)

        if new_problems and clean_up:
            passed_file = os.path.join(subdir, 'passed')
            if os.path.isfile(passed_file):
                print("Removing %s" % passed_file)
                os.remove(passed_file)

        problems += new_problems

    fmu_count = 0

    # validate the test FMUs
    for subdir, dirs, files in os.walk(os.path.join(vendor_dir, 'fmus')):

        t = segments(subdir)

        if len(t) - len(s) != 7:
            continue

        fmu_count += 1

        fmi_version, fmi_type, platform, tool_name, tool_version, model_name = t[-6:]

        if fmi_version not in ['1.0', '2.0']:
            continue

        if fmi_type not in ['cs', 'me']:
            continue

        if platform not in ['c-code', 'darwin64', 'linux32', 'linux64', 'win32', 'win64']:
            continue

        new_problems = validate_test_fmu(subdir)

        if new_problems and clean_up:
            not_compliant_file = os.path.join(subdir, 'notCompliantWithLatestRules')
            print("Adding %s" % not_compliant_file)
            with open(not_compliant_file, 'a'):
                pass

        problems += new_problems

    return fmu_count, result_count, problems


if __name__ == '__main__':

    import argparse
    import textwrap

    description = """\
    Validate cross-check results and test FMUs in vendor repositories 
    """

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=textwrap.dedent(description))

    parser.add_argument('xc_repo', default=os.getcwd(), nargs='?', help="path to the vendor repository")
    parser.add_argument('--clean-up', action='store_true', help="remove 'passed' or add 'notCompliantWithLatestRules' file")

    args = parser.parse_args()

    fmu_count, result_count, problems = validate_repo(args.xc_repo, args.clean_up)

    print()
    print("#################################")
    print("%d problems found in %s" % (len(problems), args.xc_repo))
    print("Validated %d FMUs and %d results" % (fmu_count, result_count))
    print("#################################")
    print()

    for problem in problems:
        print()
        print(problem)

    sys.exit(len(problems))



from PyQt5.QtCore import QAbstractItemModel, QModelIndex, Qt, pyqtSignal, QSortFilterProxyModel
from PyQt5.QtGui import QPixmap, QFont

from ..model_description import ScalarVariable


class TreeItem(object):

    def __init__(self, name=None, variable=None):
        self.parent = None
        self.name = name
        self.variable = variable
        self.children = []

    def find(self, name):
        for c in self.children:
            if c.name == name:
                return c
        return None

    def addChild(self, child):
        i = 0
        for i, sibling in enumerate(self.children):
            if sibling.name.upper() > child.name.upper():
                break
        self.children.insert(i, child)
        child.parent = self


class VariablesModel(QAbstractItemModel):

    COLUMN_NAMES = ['Name', 'Value Reference', 'Initial', 'Causality', 'Variability', 'Start', 'Min', 'Max', 'Unit', 'Plot', 'Description']
    COLUMN_WIDTHS = [200, 100, 70, 70, 70, 70, 70, 70, 40, 40]
    variableSelected = pyqtSignal(ScalarVariable, name='variableSelected')
    variableDeselected = pyqtSignal(ScalarVariable, name='variableDeselected')

    def __init__(self, selectedVariables, startValues, parent=None):
        super(VariablesModel, self).__init__(parent)
        self.selectedVariables = selectedVariables
        self.startValues = startValues
        self.modelDescription = None
        self.boldFont = QFont()
        self.boldFont.setBold(True)

    def columnCount(self, parent):
        return len(self.COLUMN_NAMES)

    def headerData(self, col, orientation, role):

        if role == Qt.DisplayRole:
            return self.COLUMN_NAMES[col]

        return None

    def columnData(self, v, column, role):

        if role == Qt.DecorationRole and column == 'Name':
            if v.causality in ['input', 'output']:
                return QPixmap(':/icons/%s_%s.png' % (v.type.lower(), v.causality))
            else:
                return QPixmap(':/icons/%s.png' % v.type.lower())

        elif role == Qt.CheckStateRole and column == 'Plot':
            return Qt.Checked if v in self.selectedVariables else Qt.Unchecked

        elif role == Qt.TextAlignmentRole and column in ['Value Reference', 'Start', 'Min', 'Max']:
            return Qt.AlignRight | Qt.AlignVCenter

        elif role == Qt.FontRole and column == 'Start' and v.name in self.startValues:
            return self.boldFont

        elif role == Qt.DisplayRole or role == Qt.EditRole:
            if column == 'Name':
                return v.name
            elif column == 'Value Reference':
                return v.valueReference
            elif column == 'Initial':
                return v.initial
            elif column == 'Causality':
                return v.causality
            elif column == 'Variability':
                return v.variability
            elif column == 'Start':
                if v.name in self.startValues:
                    return self.startValues[v.name]
                elif v.start is not None:
                    return str(v.start)
            elif column == 'Min':
                if v.min is not None:
                    return v.min
                elif v.declaredType is not None and v.declaredType.min is not None:
                    return v.declaredType.min
            elif column == 'Max':
                if v.max is not None:
                    return v.max
                elif v.declaredType is not None and v.declaredType.max is not None:
                    return v.declaredType.max
            elif column == 'Unit':
                if v.unit is not None:
                    return v.unit
                elif v.declaredType is not None and v.declaredType.unit is not None:
                    return v.declaredType.unit
            elif column == 'Description':
                return v.description

        return None

    def flags(self, index):

        if not index.isValid():
            return 0

        column = self.COLUMN_NAMES[index.column()]

        flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable

        if column == 'Plot':
            flags |= Qt.ItemIsUserCheckable
        elif column == 'Start':
            flags |= Qt.ItemIsEditable

        return flags

    def setData(self, index, value, role):

        if not index.isValid():
            return False

        variable = self.variableForIndex(index)
        column = self.COLUMN_NAMES[index.column()]

        if column == 'Start':
            if value:
                self.startValues[variable.name] = value
            else:
                self.startValues.pop(variable.name, None)
            return True
        elif column == 'Plot' and role == Qt.CheckStateRole:
            if value == Qt.Checked:
                self.variableSelected.emit(variable)
            else:
                self.variableDeselected.emit(variable)
            return True

        return False

    def variableForIndex(self, index):
        raise NotImplementedError()


class VariablesTableModel(VariablesModel):

    def __init__(self, selectedVariables, startValues, parent=None):
        super(VariablesTableModel, self).__init__(selectedVariables, startValues, parent)
        self.modelDescription = None

    def setModelDescription(self, modelDescription):
        self.modelDescription = modelDescription

    def data(self, index, role):
        if not index.isValid():
            return None

        column = self.COLUMN_NAMES[index.column()]

        variable = index.internalPointer()

        if role == Qt.DisplayRole and column == 'Name':
            return variable.name

        return self.columnData(variable, column, role)

    def index(self, row, column, parent):
        variable = self.modelDescription.modelVariables[row]
        return self.createIndex(row, column, variable)

        return QModelIndex()

    def parent(self, index):
        return QModelIndex()

    def rowCount(self, parent):

        if self.modelDescription is None:
            return 0

        if parent.column() > 0:
            return 0

        if not parent.isValid():
            return len(self.modelDescription.modelVariables)

        return 0

    def variableForIndex(self, index):
        return self.modelDescription.modelVariables[index.row()]


class VariablesTreeModel(VariablesModel):

    def __init__(self, selectedVariables, startValues, parent=None):
        super(VariablesTreeModel, self).__init__(selectedVariables, startValues, parent)
        self.rootItem = TreeItem()

    def setModelDescription(self, md):

        self.rootItem.children[:] = []

        for variable in md.modelVariables:

            name = variable.name

            if name.startswith('der(') and name.endswith(')'):
                prefix, name, suffix = name[:4], name[4:-1], name[-1]
            else:
                prefix = suffix = ''

            # split and remove single quotes from segments
            segments = [s[1:-1] if s.startswith("'") and s.endswith("'") else s for s in name.split('.')]

            parentItem = self.rootItem

            for segment in segments[:-1]:
                p = parentItem.find(segment)
                if p is not None:
                    parentItem = p
                else:
                    p = TreeItem(name=segment)
                    parentItem.addChild(p)
                    parentItem = p

            parentItem.addChild(TreeItem(name=prefix + segments[-1] + suffix, variable=variable))

    def data(self, index, role):

        if not index.isValid():
            return None

        column = self.COLUMN_NAMES[index.column()]

        item = index.internalPointer()

        if role == Qt.DisplayRole and column == 'Name':
            return item.name

        if item.variable is None:
            if role == Qt.DecorationRole and column == 'Name':
                return QPixmap(":/icons/subsystem.png")
            else:
                return None

        return self.columnData(item.variable, column, role)

    def index(self, row, column, parent):
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        childItem = parentItem.children[row]

        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QModelIndex()

    def parent(self, index):

        if not index.isValid():
            return QModelIndex()

        childItem = index.internalPointer()
        parentItem = childItem.parent

        if parentItem == self.rootItem:
            return QModelIndex()

        return self.createIndex(index.row(), 0, parentItem)

    def rowCount(self, parent):
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        return len(parentItem.children)

    def variableForIndex(self, index):
        item = index.internalPointer()
        return item.variable


class VariablesFilterModel(QSortFilterProxyModel):

    def __init__(self, parent=None):
        super(VariablesFilterModel, self).__init__(parent)
        self.fixedFilterString = None
        self.filterByCausality = False
        self.filterParameters = True
        self.filterCalculatedParameters = False
        self.filterInputs = True
        self.filterOutputs = True
        self.filterIndependentVariables = True
        self.filterLocalVariables = False

    def setFilterFixedString(self, pattern):
        # print('pattern: "%s"' % pattern)
        self.fixedFilterString = pattern
        self.invalidateFilter()

    def setFilterByCausality(self, filter):
        self.filterByCausality = filter
        self.invalidateFilter()

    def setFilterParameters(self, filter):
        self.filterParameters = filter
        self.invalidateFilter()

    def setFilterCalculatedParameters(self, filter):
        self.filterCalculatedParameters = filter
        self.invalidateFilter()

    def setFilterInputs(self, filter):
        self.filterInputs = filter
        self.invalidateFilter()

    def setFilterOutputs(self, filter):
        self.filterOutputs = filter
        self.invalidateFilter()

    def setFilterIndependentVariables(self, filter):
        self.filterIndependentVariables = filter
        self.invalidateFilter()

    def setFilterLocalVariables(self, filter):
        self.filterLocalVariables = filter
        self.invalidateFilter()

    def filterAcceptsRow(self, source_row, source_parent):

        index = self.sourceModel().index(source_row, 0, source_parent)
        item = index.internalPointer()

        if isinstance(item, TreeItem):
            if item.variable is None:
                return True
            else:
                variable = item.variable
        else:
            variable = item

        if self.filterByCausality:

            if variable.causality == 'parameter':
                if not self.filterParameters:
                    return False
            elif variable.causality == 'calculatedParameter':
                if not self.filterCalculatedParameters:
                    return False
            elif variable.causality == 'input':
                if not self.filterInputs:
                    return False
            elif variable.causality == 'output':
                if not self.filterOutputs:
                    return False
            elif variable.causality == 'independent':
                if not self.filterIndependentVariables:
                    return False
            else:
                if not self.filterLocalVariables:
                    return False

        if self.fixedFilterString:
            return self.fixedFilterString in variable.name
        else:
            return True
        # return super(VariablesFilterModel, self).filterAcceptsColumn(source_row, source_parent)



""" Simulate Functional Mock-up Units (FMUs) in Python """

import sys
import os
from ctypes import *
import _ctypes
from typing import Union, IO, List

__version__ = '0.3.1'

# experimental
plot_library = 'matplotlib'  # 'plotly'


# determine the platform
if sys.platform.startswith('win'):
    platform = 'win'
    system = 'windows'
    sharedLibraryExtension = '.dll'
elif sys.platform.startswith('linux'):
    platform = 'linux'
    system = 'linux'
    sharedLibraryExtension = '.so'
elif sys.platform.startswith('darwin'):
    platform = 'darwin'
    system = 'darwin'
    sharedLibraryExtension = '.dylib'
else:
    raise Exception("Unsupported platform: " + sys.platform)


# load the C library functions
if sys.platform.startswith('win'):
    calloc = cdll.msvcrt.calloc
    realloc = cdll.msvcrt.realloc
    free = cdll.msvcrt.free
    freeLibrary = _ctypes.FreeLibrary
else:
    from ctypes.util import find_library
    libc = CDLL(find_library("c"))
    calloc = libc.calloc
    realloc = libc.realloc
    free = libc.free
    freeLibrary = _ctypes.dlclose

calloc.argtypes = [c_size_t, c_size_t]
calloc.restype = c_void_p

realloc.argtypes = [c_void_p, c_size_t]
realloc.restype = c_void_p

free.argtypes = [c_void_p]
free.restype = None

if sys.maxsize > 2**32:
    platform += '64'
    architecture = 'x86_64'
else:
    platform += '32'
    architecture = 'i686'

platform_tuple = architecture + '-' + system


def supported_platforms(filename: Union[str, IO]):
    """ Get the platforms supported by the FMU without extracting it

    Parameters:
        filename    filename of the FMU, directory with extracted FMU or file like object

    Returns:
        platforms   a list of supported platforms supported by the FMU
    """

    # get the files within the FMU
    if isinstance(filename, str) and os.path.isdir(filename):  # extracted FMU
        names = []
        for dirpath, _, filenames in os.walk(filename):
            for name in filenames:
                abspath = os.path.join(dirpath, name)
                names.append(os.path.relpath(abspath, start=filename).replace('\\', '/'))
    else:  # FMU as path or file like object
        import zipfile
        with zipfile.ZipFile(filename, 'r') as zf:
            names = zf.namelist()

    platforms = []

    # check for the C-sources
    for name in names:
        head, tail = os.path.split(name)
        if head == 'sources' and tail.endswith('.c'):
            platforms.append('c-code')
            break

    # check for *.dylib on Mac
    for name in names:
        head, tail = os.path.split(name)
        if head in {'binaries/darwin64', 'binaries/x86_64-darwin'} and tail.endswith('.dylib'):
            platforms.append('darwin64')
            break

    # check for *.so on Linux
    for bitness, architecture in [('32', 'i686'), ('64', 'x86_64')]:
        for name in names:
            head, tail = os.path.split(name)
            if head in {'binaries/linux' + bitness, 'binaries/' + architecture + '-linux'} and tail.endswith('.so'):
                platforms.append('linux' + bitness)
                break

    # check for *.dll on Windows
    for bitness, architecture in [('32', 'i686'), ('64', 'x86_64')]:
        for name in names:
            head, tail = os.path.split(name)
            if head in {'binaries/win' + bitness, 'binaries/' + architecture + '-windows'} and tail.endswith('.dll'):
                platforms.append('win' + bitness)
                break

    return platforms


def fmi_info(filename):
    """ Read the FMI version and supported FMI types from the FMU without extracting it

    Parameters:
        filename      filename of the FMU

    Returns:
        fmi_version   FMI version as a string ('1.0' or '2.0')
        fmi_types     list of supported FMI types ('CoSimulation', 'ModelExchange')
    """

    from lxml import etree
    import zipfile

    fmi_types = []

    # open the FMU
    with zipfile.ZipFile(filename, 'r') as zf:

        # read the model description
        md = zf.open('modelDescription.xml')
        tree = etree.parse(md)
        root = tree.getroot()
        fmi_version = root.get('fmiVersion')

        # get the supported FMI types
        if fmi_version == '1.0':

            if root.find('Implementation') is not None:
                fmi_types.append('CoSimulation')
            else:
                fmi_types.append('ModelExchange')

        elif fmi_version == '2.0':

            if root.find('ModelExchange') is not None:
                fmi_types.append('ModelExchange')

            if root.find('CoSimulation') is not None:
                fmi_types.append('CoSimulation')

        else:
            raise Exception("Unsupported FMI version %s" % fmi_version)

    return fmi_version, fmi_types


def extract(filename, unzipdir=None, include=None):
    """ Extract a ZIP archive to a temporary directory

    Parameters:
        filename    filename of the ZIP archive
        unzipdir    target directory (None: create temporary directory)
        include     a filter function to select the files to extract
    Returns:
        unzipdir    path to the directory that contains the extracted files
    """

    from tempfile import mkdtemp
    import zipfile

    if unzipdir is None:
        unzipdir = mkdtemp()

    # expand the 8.3 paths on windows
    if sys.platform.startswith('win') and '~' in unzipdir:
        import win32file
        unzipdir = win32file.GetLongPathName(unzipdir)

    with zipfile.ZipFile(filename, 'r') as zf:

        # check filenames
        for name in zf.namelist():
            
            if '\\' in name:
                raise Exception("Illegal path %s found in %s. All slashes must be forward slashes." % (name, filename))

            if ':' in name or name.startswith('/'):
                raise Exception("Illegal path %s found in %s. The path must not contain a drive or device letter, or a leading slash." % (name, filename))

        members = filter(include, zf.namelist()) if include else None

        # extract the archive
        zf.extractall(unzipdir, members=members)

    return unzipdir


def dump(filename: Union[str, IO], causalities: List[str] = ['input', 'output']):
    """ Print the model information and variables of an FMU

    Parameters:
        filename     filename of the FMU
        causalities  the causalities of the variables to include
    """

    from .util import fmu_info
    print(fmu_info(filename=filename, causalities=causalities))


# make the functions available in the fmpy module
from .model_description import read_model_description
from .simulation import simulate_fmu, instantiate_fmu
from .util import plot_result, read_csv, write_csv



import os
from typing import List, IO, Union

import fmpy
import numpy as np


def read_csv(filename, variable_names=[], validate=True, structured=False):
    """ Read a CSV file that conforms to the FMI cross-check rules

    Parameters:
        filename         name of the CSV file to read
        variable_names   list of variables to read (default: read all)
        structured       convert structured names to arrays

    Returns:
        traj             the trajectories read from the CSV file
    """

    # pass an empty string as deletechars to preserve special characters
    traj = np.genfromtxt(filename, delimiter=',', names=True, deletechars='')

    if structured:
        arrays = {}

        cols = []
        traj_ = []

        for name, type_ in traj.dtype.descr:
            if name.endswith(']'):
                i = name.rfind('[')
                basename = name[:i]
                if basename not in arrays:
                    arrays[basename] = []
                arrays[basename].append((int(name[i + 1:-1]) - 1, traj[name]))
            else:
                cols.append((name, type_))
                traj_.append(traj[name].tolist())

        for name, value in arrays.items():
            subs, arrs = zip(*value)
            cols.append((name, '<f8', (max(subs) + 1,)))
            traj_.append(list(zip(*arrs)))

        traj = np.array(list(zip(*traj_)), dtype=np.dtype(cols))

    if not validate:
        return traj

    # get the time
    time = traj[traj.dtype.names[0]]

    # check if the time is monotonically increasing
    if traj.size > 1 and np.any(np.diff(time) < 0):
        raise Exception("Values in first column (time) are not monotonically increasing")

    # get the trajectory names (without the time)
    traj_names = traj.dtype.names[1:]

    # check if the variable names match the trajectory names
    for variable_name in variable_names:
        if variable_name not in traj_names:
            raise Exception("Trajectory of '" + variable_name + "' is missing")

    return traj


def write_csv(filename, result, columns=None):
    """ Save a simulation result as a CSV file

    Parameters:
        filename   name of the CSV file to write
        result     structured NumPy array that holds the result
        columns    list of column names to save (None: save all)
    """

    if columns is not None:
        result = result[['time'] + columns]

    # serialize multi-dimensional signals
    cols = []
    data = []

    for name in result.dtype.names:
        dtype = result.dtype[name]
        if len(dtype.shape) > 0:
            subtype = dtype.subdtype[0].type
            y = result[name]
            for i in np.ndindex(dtype.shape):
                # convert index to 1-based subscripts
                subs = ','.join(map(lambda sub: str(sub + 1), i))
                cols.append(('%s[%s]' % (name, subs), subtype))
                sl = tuple([slice(0, None)] + [slice(s, s + 1) for s in i])
                data.append(y[sl].flatten())
        else:
            cols.append((name, dtype.type))
            data.append(result[name])

    result = np.array(list(zip(*data)), dtype=np.dtype(cols))

    header = ','.join(map(lambda s: '"' + s + '"', result.dtype.names))
    np.savetxt(filename, result, delimiter=',', header=header, comments='', fmt='%g')


def read_ref_opt_file(filename):

    opts = {}

    with open(filename, 'r') as f:
        for line in f:
            segments = line.split(',')
            if len(segments) == 2:
                opts[segments[0]] = float(segments[1])

    # check for required elements
    for element in ['StepSize', 'StartTime', 'StopTime', 'RelTol']:
        if not element in opts:
            raise Exception("Missing element '%s'" % element)

    start_time = opts['StartTime']
    stop_time = opts['StopTime']
    step_size = opts['StepSize']

    if start_time >= stop_time:
        raise Exception("StartTime must be < StopTime")

    if step_size < 0 or step_size > (stop_time - start_time):
        raise Exception("StepSize must be >= 0 and <= (StopTime - StartTime)")

    return opts


def validate_signal(t, y, t_ref, y_ref, num=1000, dx=20, dy=0.1):
    """ Validate a signal y(t) against a reference signal y_ref(t_ref) by creating a band
    around y_ref and finding the values in y outside the band

    Parameters:

        t       time of the signal
        y       values of the signal
        t_ref   time of the reference signal
        y_ref   values of the reference signal
        num     number of samples for the band
        dx      horizontal width of the band in samples
        dy      vertical distance of the band to y_ref

    Returns:

        t_band  time values of the band
        y_min   lower limit of the band
        y_max   upper limit of the band
        i_out   indices of the values in y outside the band
    """

    from scipy.ndimage.filters import maximum_filter1d, minimum_filter1d
    from scipy.interpolate import interp1d

    # re-sample the reference signal into a uniform grid
    t_band = np.linspace(start=t_ref[0], stop=t_ref[-1], num=num)

    # make t_ref strictly monotonic by adding epsilon to duplicate sample times
    for i in range(1, len(t_ref)):
        while t_ref[i - 1] >= t_ref[i]:
            t_ref[i] = t_ref[i] + 1e-13

    interp_method = 'linear' if y.dtype == np.float64 else 'zero'
    y_band = interp1d(x=t_ref, y=y_ref, kind=interp_method)(t_band)

    y_band_min = np.min(y_band)
    y_band_max = np.max(y_band)

    # calculate the width of the band
    if y_band_min == y_band_max:
        w = 0.5 if y_band_min == 0 else np.abs(y_band_min) * dy
    else:
        w = (y_band_max - y_band_min) * dy

    # calculate the lower and upper limits
    y_min = minimum_filter1d(input=y_band, size=dx) - w
    y_max = maximum_filter1d(input=y_band, size=dx) + w

    # find outliers
    y_min_i = np.interp(x=t, xp=t_band, fp=y_min)
    y_max_i = np.interp(x=t, xp=t_band, fp=y_max)
    i_out = np.logical_or(y < y_min_i, y > y_max_i)

    # do not count outliers outside the t_ref
    i_out = np.logical_and(i_out, t > t_band[0])
    i_out = np.logical_and(i_out, t < t_band[-1])

    return t_band, y_min, y_max, i_out


def validate_result(result, reference, stop_time=None):
    """ Validate a simulation result against a reference result

    Parameters:
        result      structured NumPy array where the first column is the time
        reference   same as result

    Returns:
        rel_out     the largest relative deviation of all signals
    """

    t_ref = reference[reference.dtype.names[0]]
    t_res = result[result.dtype.names[0]]

    # at least two samples are required
    if result.size < 2:
        return 1

    # check if stop time has been reached
    if stop_time is not None and t_res[-1] < stop_time:
        return 1

    rel_out = 0

    # find the signal with the most outliers
    for name in result.dtype.names[1:]:

        if name not in reference.dtype.names:
            continue

        y_ref = reference[name]
        y_res = result[name]
        _, _, _, outliers = validate_signal(t=t_res, y=y_res, t_ref=t_ref, y_ref=y_ref)
        rel_out = np.max([np.sum(outliers) / float(len(outliers)), rel_out])

    return rel_out


def create_plotly_figure(result, names=None, events=False, time_unit=None):

    import plotly.graph_objects as go
    from plotly.subplots import make_subplots

    model_description = getattr(result, 'modelDescription', None)

    units = {}
    display_units = {}

    if model_description:
        # get the units and display units
        for unit in model_description.unitDefinitions:
            if unit.displayUnits:
                display_units[unit.name] = unit.displayUnits[0]

        for v in model_description.modelVariables:
            unit = v.unit
            if unit is None and v.declaredType:
                unit = v.declaredType.unit
            units[v.name] = unit

    time = result['time']

    min_ticks = 5

    if time_unit is None:

        time_span = time[-1] - time[0]

        if time_span < min_ticks * 1e-6:
            time_unit = 'ns'
        elif time_span < min_ticks * 1e-3:
            time_unit = 'us'
        elif time_span < 1:
            time_unit = 'ms'
        elif time_span > min_ticks * 365 * 24 * 60 * 60:
            time_unit = 'years'
        elif time_span > min_ticks * 24 * 60 * 60:
            time_unit = 'days'
        elif time_span > min_ticks * 60 * 60:
            time_unit = 'h'
        elif time_span > min_ticks * 60:
            time_unit = 'min'
        else:
            time_unit = 's'

    if time_unit == 'ns':
        time *= 1e9
    elif time_unit == 'us':
        time *= 1e6
    elif time_unit == 'ms':
        time *= 1e3
    elif time_unit == 's':
        pass
    elif time_unit == 'min':
        time /= 60
    elif time_unit == 'h':
        time /= 60 * 60
    elif time_unit == 'days':
        time /= 24 * 60 * 60
    elif time_unit == 'years':
        time /= 365 * 24 * 60 * 60
    else:
        raise Exception(f'time_unit must be one of "ns", "us", "ms", "s", "min", "h", "days" or "years" but was "{time_unit}".')

    if names is None:
        # plot at most 20 signals
        names = result.dtype.names[1:20]

    fig = make_subplots(rows=len(names), cols=1, shared_xaxes=True)

    for i, name in enumerate(names):

        y = result[name]
        unit = units.get(name)

        if unit in display_units:
            display_unit = display_units[unit]
            y = y * display_unit.factor + display_unit.offset
            unit = display_unit.name

        line = dict(color='#636efa', width=1)

        if y.dtype in [np.float32, np.float64]:
            trace = go.Scatter(x=time, y=y, name=name, line=line)
        elif y.dtype == bool:
            trace = go.Scatter(x=time, y=y.astype(int), name=name, line=line, fill='tozeroy', fillcolor='rgba(0,0,255,0.1)', line_shape='hv')
            fig['layout'][f'yaxis{i + 1}'].update(tickvals=[0, 1], ticktext=['false', 'true'], range=[-0.1, 1.1], fixedrange=True)
        else:
            trace = go.Scatter(x=time, y=y, name=name, line=line, line_shape='hv')

        fig.add_trace(trace, row=i + 1, col=1)

        fig['layout'][f'yaxis{i + 1}'].update(title=f"{name} [{unit}]" if unit else name)

    if events:
        for t_event in time[np.argwhere(np.diff(time) == 0).flatten()]:
            fig.add_vline(x=t_event, line={'color': '#fbe424', 'width': 1})

    fig['layout']['height'] = 160 * len(names) + 30 * max(0, 5 - len(names))
    fig['layout']['margin']['t'] = 30
    fig['layout']['margin']['b'] = 0
    fig['layout']['margin']['r'] = 30
    fig['layout']['plot_bgcolor'] = 'rgba(0,0,0,0)'
    fig['layout'][f'xaxis{len(names)}'].update(title=f'time [{time_unit}]')

    axes_attrs = dict(showgrid=True, gridwidth=1, ticklen=0, gridcolor='LightGrey', linecolor='black', showline=True, zerolinewidth=1, zerolinecolor='LightGrey')
    fig.update_xaxes(zeroline=True, **axes_attrs)
    fig.update_yaxes(**axes_attrs)

    fig.update_layout(showlegend=False)

    return fig


def plot_result(result, reference=None, names=None, filename=None, window_title=None, events=False):
    """ Plot a collection of time series.

    Parameters:
        result:       structured NumPy Array that contains the time series to plot where 'time' is the independent variable
        reference:    optional reference signals with the same structure as `result`
        names:        variables to plot
        filename:     when provided the plot is saved as `filename` instead of showing the figure
        window_title: title for the figure window
        events:       draw vertical lines at events
    """

    from . import plot_library

    if plot_library == 'plotly':
        figure = create_plotly_figure(result, names=names, events=events)
        if filename is None:
            figure.show()
        else:
            figure.write_image(filename)
        return

    import matplotlib.pylab as pylab
    import matplotlib.pyplot as plt
    import matplotlib.transforms as mtransforms
    from matplotlib.ticker import MaxNLocator
    from collections.abc import Iterable

    params = {
        'legend.fontsize': 8,
        'axes.labelsize': 8,
        'xtick.labelsize': 8,
        'ytick.labelsize': 8,
        'axes.linewidth': 0.5,
    }

    pylab.rcParams.update(params)

    time = result['time']

    if names is None:
        # plot at most 20 signals
        names = result.dtype.names[1:20]

    if len(names) > 0:

        # indent label 0.015 inch / character
        label_x = -0.015 * np.max(list(map(len, names)) + [8])

        fig, axes = plt.subplots(len(names), sharex=True)

        fig.set_facecolor('white')

        if not isinstance(axes, Iterable):
            axes = [axes]

        if events:
            t_event = time[np.argwhere(np.diff(time) == 0)]

        for ax, name in zip(axes, names):

            y = result[name]

            ax.grid(b=True, which='both', color='0.8', linestyle='-', zorder=0)

            ax.tick_params(direction='in')

            if events:
                for t in t_event:
                    ax.axvline(x=t, color='y', linewidth=1)

            if reference is not None and name in reference.dtype.names:
                t_ref = reference[reference.dtype.names[0]]
                y_ref = reference[name]

                t_band, y_min, y_max, i_out = validate_signal(t=time, y=y, t_ref=t_ref, y_ref=y_ref)

                ax.fill_between(t_band, y_min, y_max, facecolor=(0, 0.5, 0), alpha=0.1)
                ax.plot(t_band, y_min, color=(0, 0.5, 0), linewidth=1, label='lower bound', zorder=101, alpha=0.5)
                ax.plot(t_band, y_max, color=(0, 0.5, 0), linewidth=1, label='upper bound', zorder=101, alpha=0.5)

                # mark the outliers
                # use the data coordinates for the x-axis and the axes coordinates for the y-axis
                trans = mtransforms.blended_transform_factory(ax.transData, ax.transAxes)
                ax.fill_between(time, 0, 1, where=i_out, facecolor='red', alpha=0.5, transform=trans)

            if y.dtype == np.float64:
                # find left indices of discontinuities
                i_disc = np.flatnonzero(np.diff(time) == 0)
                i_disc = np.append(i_disc + 1, len(time))
                i0 = 0
                for i1 in i_disc:
                    ax.plot(time[i0:i1], y[i0:i1], color='b', linewidth=0.9, label='result', zorder=101)
                    i0 = i1
            else:
                ax.hlines(y[:-1], time[:-1], time[1:], colors='b', linewidth=1, label='result', zorder=101)
                ax.yaxis.set_major_locator(MaxNLocator(integer=True))

            if y.dtype == bool:
                # use fixed range and labels and fill area
                ax.set_ylim(-0.25, 1.25)
                ax.yaxis.set_ticks([0, 1])
                ax.yaxis.set_ticklabels(['false', 'true'])
                if y.ndim == 1:
                    ax.fill_between(time, y, 0, step='post', facecolor='b', alpha=0.1)
            else:
                ax.margins(x=0, y=0.05)

            if time.size < 200:
                ax.scatter(time, y, color='b', s=5, zorder=101)

            ax.set_ylabel(name, horizontalalignment='left', rotation=0)

            # align the y-labels
            ax.get_yaxis().set_label_coords(label_x, 0.5)

        # set the window title
        if window_title is not None:
            fig.canvas.set_window_title(window_title)

        def onresize(event):
            fig = plt.gcf()

            w = fig.get_figwidth()

            # tight_layout() crashes on very small figures
            if w < 3:
                return

            x = label_x * (8.0 / w)

            # update label coordinates
            for ax in fig.get_axes():
                ax.get_yaxis().set_label_coords(x, 0.5)

            # update layout
            plt.tight_layout()

        # update layout when the plot is re-sized
        fig.canvas.mpl_connect('resize_event', onresize)

        fig.set_size_inches(w=8, h=1.5 * len(names), forward=True)

        plt.tight_layout()

        if filename is None:
            plt.show()
        else:
            dir, _ = os.path.split(filename)
            if not os.path.isdir(dir):
                os.makedirs(dir)
            fig.savefig(filename)
            plt.close(fig)


def fmu_path_info(path):
    """ Extract information from the FMU's file path """
    
    head = path
    values = []

    while True:
        head, tail = os.path.split(head)

        if not tail:
            break

        values.append(tail)

        if tail == 'FMI_1.0' or tail == 'FMI_2.0':
            break

    keys = ['model_name', 'tool_version', 'tool_name', 'platform', 'fmi_type', 'fmi_version']

    return dict(zip(keys, values))


def sha256_checksum(filename):
    """ Create a SHA256 checksum form a file """

    import hashlib

    sha256 = hashlib.sha256()

    with open(filename, 'rb') as f:
        for block in iter(lambda: f.read(65536), b''):
            sha256.update(block)

    return sha256.hexdigest()


def download_file(url: str, checksum: str = None) -> str:
    """ Download a file to the current directory

        returns the filename of the downloaded file
    """

    filename = os.path.basename(url)

    if checksum is not None and os.path.isfile(filename):
        hash = sha256_checksum(filename)
        if hash.startswith(checksum.lower()):
            return  # file already exists

    import requests

    print('Downloading ' + url)

    status_code = -1

    # try to download the file three times
    try:
        for _ in range(3):
            if status_code != 200:
                response = requests.get(url)
                status_code = response.status_code
    except:
        pass  # ignore

    if status_code != 200:
        raise Exception("Failed to download %s (status code: %d)" % (url, status_code))

    # write the file
    with open(filename, 'wb') as f:
        f.write(response.content)

    if checksum is not None:
        hash = sha256_checksum(filename)
        if not hash.startswith(checksum):
            raise Exception("%s has the wrong SHA256 checksum. Expected %s but was %s." % (filename, checksum, hash))

    return filename


def download_test_file(fmi_version, fmi_type, tool_name, tool_version, model_name, filename):
    """ Download a file from the Test FMUs repository to the current directory """

    from . import platform

    # for backwards compatibility
    if fmi_type == 'ModelExchange':
        fmi_type = 'me'
    elif fmi_type == 'CoSimulation':
        fmi_type = 'cs'

    # build the URL
    url = 'https://github.com/modelica/fmi-cross-check/raw/master/fmus'
    url = '/'.join([url, fmi_version, fmi_type, platform, tool_name, tool_version, model_name, filename])

    download_file(url)


def fmu_info(filename: Union[str, IO], causalities: List[str] = ['input', 'output']) -> str:
    """ Dump the info for an FMU

    Parameters:
        filename     filename of the FMU
        causalities  the causalities of the variables to include

    Returns the info as a multi line string
    """

    from .model_description import read_model_description
    from . import supported_platforms

    md = read_model_description(filename, validate=False)
    platforms = supported_platforms(filename)

    fmi_types = []
    if md.modelExchange is not None:
        fmi_types.append('Model Exchange')
    if md.coSimulation is not None:
        fmi_types.append('Co-Simulation')

    l = [f"""
Model Info

  FMI Version        {md.fmiVersion}
  FMI Type           {', '.join(fmi_types)}
  Model Name         {md.modelName}
  Description        {md.description}
  Platforms          {', '.join(platforms)}
  Continuous States  {md.numberOfContinuousStates}
  Event Indicators   {md.numberOfEventIndicators}
  Variables          {len(md.modelVariables)}
  Generation Tool    {md.generationTool}
  Generation Date    {md.generationDateAndTime}
"""]

    if md.defaultExperiment:

        ex = md.defaultExperiment

        l.append('Default Experiment')
        l.append("")
        if ex.startTime:
            l.append(f"  Start Time         {ex.startTime}")
        if ex.stopTime:
            l.append(f"  Stop Time          {ex.stopTime}")
        if ex.tolerance:
            l.append(f"  Tolerance          {ex.tolerance}")
        if ex.stepSize:
            l.append(f"  Step Size          {ex.stepSize}")

    inputs = []
    outputs = []

    for v in md.modelVariables:
        if v.causality == 'input':
            inputs.append(v.name)
        if v.causality == 'output':
            outputs.append(v.name)

    l.append("")
    l.append("Variables (%s)" % ', '.join(causalities))
    l.append("")
    l.append('  Name               Causality              Start Value  Unit     Description')
    for v in md.modelVariables:
        if v.causality not in causalities:
            continue

        start = str(v.start) if v.start is not None else ''

        unit = v.declaredType.unit if v.declaredType else v.unit

        args = ['' if s is None else str(s) for s in [v.name, v.causality, start, unit, v.description]]

        l.append('  {:18} {:10} {:>23}  {:8} {}'.format(*args))

    return '\n'.join(l)


def visual_studio_installation_path():
    """ Get the VisualStudio 2017 installation path """

    try:
        from subprocess import Popen, PIPE
        import os
        vswhere = '"' + os.environ['ProgramFiles(x86)'] + r'\Microsoft Visual Studio\Installer\vswhere.exe"'
        command = vswhere +  ' -latest -products * -requires Microsoft.Component.MSBuild -property installationPath'
        proc = Popen(command, stdout=PIPE)
        output, _ = proc.communicate()
        return output.decode('utf-8').strip()
    except Exception as e:
        pass  # do noting

    return None


def visual_c_versions():
    """ Detect installed Visual C compilers

    Returns: a sorted list of detected Visual C versions e.g. [90, 120, 140]
    """

    versions = []

    # up to Visual Studio 2015
    for key in os.environ.keys():
        if key.upper().startswith('VS') and key.upper().endswith('COMNTOOLS'):
            versions.append(int(key[len('VS'):-len('COMNTOOLS')]))

    # Visual Studio 2017
    installation_path = visual_studio_installation_path()

    if installation_path is not None:
        if '2017' in installation_path:
            versions.append(150)
        if '2019' in installation_path:
            versions.append(160)

    return sorted(versions)


def compile_dll(model_description, sources_dir, compiler=None, target_platform=None):
    """ Compile the shared library

    Parameters:
        sources_dir    directory that contains the FMU's source code
        compiler       compiler to use (None: use Visual C on Windows, GCC otherwise)
        platform       platform to compile for the binary for
    """

    from . import platform, system
    import subprocess

    if target_platform is None:
        target_platform = platform

    if model_description.fmiVersion == '1.0':
        raise Exception("FMI 1.0 source FMUs are currently not supported")

    if compiler is None:
        if target_platform in ['win32', 'win64', 'x86_64-windows']:
            compiler = 'vc'
        else:
            compiler = 'gcc'

    include_dir = os.path.join(os.path.dirname(__file__), 'c-code')
    preprocessor_definitions = []

    source_files = []

    if len(model_description.buildConfigurations) == 0:
        raise Exception("No build configuration found.")

    build_configuration = model_description.buildConfigurations[0]

    if len(build_configuration.sourceFileSets) > 1:
        raise Exception("More than one SourceFileSet is not supported.")

    source_file_set = build_configuration.sourceFileSets[0]

    source_files += source_file_set.sourceFiles

    for definition in source_file_set.preprocessorDefinitions:
        literal = definition.name
        if definition.value is not None:
            literal += '=' + definition.value
        preprocessor_definitions.append(literal)

    if len(source_files) == 0:
        raise Exception("No source files specified in the model description.")

    print('Compiling platform binary...')

    sources = ' '.join(source_files)
    model_identifier = build_configuration.modelIdentifier

    if compiler == 'vc':

        target = model_identifier + '.dll'

        vc_versions = visual_c_versions()

        toolset = 'x86_amd64' if target_platform in ['win64', 'x86_64-windows'] else 'x86'

        definitions = ' '.join(f' /D{d}' for d in preprocessor_definitions)

        if len(vc_versions) == 0:
            raise Exception("No VisualStudio found")

        # use the latest version
        vc_version = vc_versions[-1]

        if vc_version < 150:
            command = rf'call "%%VS{vc_version}COMNTOOLS%%\..\..\VC\vcvarsall.bat" {toolset}'
        else:
            installation_path = visual_studio_installation_path()
            command = rf'call "{installation_path}\VC\Auxiliary\Build\vcvarsall.bat" {toolset}'

        command += f' && cl /LD /I. /I"{include_dir}" {definitions} /Fe{model_identifier} shlwapi.lib {sources}'

    elif compiler == 'gcc':

        definitions = ' '.join(f' -D{d}' for d in preprocessor_definitions)

        if target_platform in ['linux64', 'x86_64-linux']:

            cc = 'gcc'
            target = model_identifier + '.so'

            if system == 'windows':
                cc = 'wsl ' + cc
                output = subprocess.check_output(f"wsl wslpath -a '{include_dir}'")
                include_dir = output.decode('utf-8').strip()

            command = f'{cc} -c -I. -I{include_dir} -fPIC {definitions} {sources}'
            command += f' && {cc} -static-libgcc -shared -o{target} *.o -lm'

        elif target_platform in ['darwin64', 'x86_64-darwin']:

            target = model_identifier + '.dylib'

            command = f'gcc -c -I. -I{include_dir} {definitions} {sources}'
            command += f' && gcc -shared -o{target} *.o -lm'

        else:
            raise Exception("Unsupported target platform for selected compiler: '%s'" % compiler)
    else:
        raise Exception("Unsupported compiler: '%s'" % compiler)

    cur_dir = os.getcwd()
    os.chdir(sources_dir)
    status = os.system(command)
    os.chdir(cur_dir)

    print(sources_dir)
    print(command)

    dll_path = os.path.join(sources_dir, target)

    if status != 0 or not os.path.isfile(dll_path):
        raise Exception('Failed to compile shared library')

    return str(dll_path)


def compile_platform_binary(filename, output_filename=None, target_platform=None):
    """ Compile the binary of an FMU for the current platform and add it to the FMU

    Parameters:
        filename:         filename of the source code FMU
        output_filename:  filename of the FMU with the compiled binary (None: overwrite existing FMU)
    """

    from . import read_model_description, extract, platform, platform_tuple
    import zipfile
    from shutil import copyfile, rmtree

    unzipdir = extract(filename)

    model_description = read_model_description(filename)

    if target_platform is None:
        target_platform = platform if model_description.fmiVersion in ['1.0', '2.0'] else platform_tuple

    binary = compile_dll(model_description=model_description,
                         sources_dir=os.path.join(unzipdir, 'sources'),
                         target_platform=target_platform)

    unzipdir2 = extract(filename)

    target_platform_dir = os.path.join(unzipdir2, 'binaries', target_platform)

    if not os.path.exists(target_platform_dir):
        os.makedirs(target_platform_dir)

    copyfile(src=binary, dst=os.path.join(target_platform_dir, os.path.basename(binary)))

    if output_filename is None:
        output_filename = filename  # overwrite the existing archive

    # create a new archive from the existing files + compiled binary
    with zipfile.ZipFile(output_filename, 'w', zipfile.ZIP_DEFLATED) as zf:
        base_path = os.path.normpath(unzipdir2)
        for dirpath, dirnames, filenames in os.walk(unzipdir2):
            for name in sorted(dirnames):
                path = os.path.normpath(os.path.join(dirpath, name))
                zf.write(path, os.path.relpath(path, base_path))
            for name in filenames:
                path = os.path.normpath(os.path.join(dirpath, name))
                if os.path.isfile(path):
                    zf.write(path, os.path.relpath(path, base_path))

    # clean up
    rmtree(unzipdir, ignore_errors=True)
    rmtree(unzipdir2, ignore_errors=True)


def add_remoting(filename, host_platform, remote_platform):

    from . import extract, read_model_description, supported_platforms
    from shutil import copyfile, rmtree
    import zipfile
    import os

    platforms = supported_platforms(filename)

    if host_platform == 'win64' and remote_platform == 'win32':

        if 'win64' in platforms:
            raise Exception('The FMU already supports "win64".')

        if 'win32' not in platforms:
            raise Exception('The FMU does not support the platform "win32".')

    elif host_platform == 'linux64' and remote_platform == 'win64':

        if 'linux64' in platforms:
            raise Exception('The FMU already supports "linux64".')

        if 'win64' not in platforms:
            raise Exception('The FMU does not support the platform "win64".')
    else:

        raise Exception("Remoting is not supported for the given combination of host and remote platform.")

    model_description = read_model_description(filename)

    current_dir = os.path.dirname(__file__)

    if host_platform == 'win64':
        client = os.path.join(current_dir, 'remoting', 'win64', 'client.dll')
        server = os.path.join(current_dir, 'remoting', 'win32', 'server.exe')
    else:
        client = os.path.join(current_dir, 'remoting', 'linux64', 'client.so')
        server = os.path.join(current_dir, 'remoting', 'win64', 'server.exe')

    license = os.path.join(current_dir, 'remoting', 'license.txt')

    tempdir = extract(filename)

    if model_description.coSimulation:
        model_identifier = model_description.coSimulation.modelIdentifier
    else:
        model_identifier = model_description.modelExchange.modelIdentifier

    # copy the binaries & license
    if host_platform == 'win64':
        os.mkdir(os.path.join(tempdir, 'binaries', 'win64'))
        copyfile(client, os.path.join(tempdir, 'binaries', 'win64', model_identifier + '.dll'))
        copyfile(server, os.path.join(tempdir, 'binaries', 'win32', 'server.exe'))
    else:
        os.mkdir(os.path.join(tempdir, 'binaries', 'linux64'))
        copyfile(client, os.path.join(tempdir, 'binaries', 'linux64', model_identifier + '.so'))
        copyfile(server, os.path.join(tempdir, 'binaries', 'win64', 'server.exe'))

    licenses_dir = os.path.join(tempdir, 'documentation', 'licenses')
    os.makedirs(licenses_dir, exist_ok=True)
    copyfile(license, os.path.join(tempdir, 'documentation', 'licenses', 'fmpy-remoting-binaries.txt'))

    # create a new archive from the existing files + remoting binaries
    with zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED) as zf:
        base_path = os.path.normpath(tempdir)
        for dirpath, dirnames, filenames in os.walk(tempdir):
            for name in sorted(dirnames):
                path = os.path.normpath(os.path.join(dirpath, name))
                zf.write(path, os.path.relpath(path, base_path))
            for name in filenames:
                path = os.path.normpath(os.path.join(dirpath, name))
                if os.path.isfile(path):
                    zf.write(path, os.path.relpath(path, base_path))

    # clean up
    rmtree(tempdir, ignore_errors=True)


def auto_interval(t):
    """ Find a nice interval that divides t into 500 - 1000 steps """

    h = 10 ** (np.round(np.log10(t)) - 3)

    n_samples = t / h

    if n_samples >= 2500:
        h *= 5
    elif n_samples >= 2000:
        h *= 4
    elif n_samples >= 1000:
        h *= 2
    elif n_samples <= 200:
        h /= 5
    elif n_samples <= 250:
        h /= 4
    elif n_samples <= 500:
        h /= 2

    return h


def change_fmu(input_file, output_file=None, start_values={}):
    """ Make changes to an FMU """

    from lxml import etree
    import zipfile
    from fmpy import extract
    from shutil import rmtree

    if output_file is None:
        output_file = input_file

    tempdir = extract(input_file)

    # read the model description
    with zipfile.ZipFile(input_file, 'r') as zf:
        xml = zf.open('modelDescription.xml')
        tree = etree.parse(xml)

    root = tree.getroot()

    # apply the start values
    for variable in root.find('ModelVariables'):
        if variable.get("name") in start_values:
            for child in variable.getchildren():
                if child.tag in ['Real', 'Integer', 'Enumeration', 'Boolean', 'String']:
                    child.set('start', start_values[variable.get("name")])

    # write the new model description
    tree.write(os.path.join(tempdir, 'modelDescription.xml'))

    # create a new archive from the modified files
    with zipfile.ZipFile(output_file, 'w', zipfile.ZIP_DEFLATED) as zf:
        base_path = os.path.normpath(tempdir)
        for dirpath, dirnames, filenames in os.walk(tempdir):
            for name in sorted(dirnames):
                path = os.path.normpath(os.path.join(dirpath, name))
                zf.write(path, os.path.relpath(path, base_path))
            for name in filenames:
                path = os.path.normpath(os.path.join(dirpath, name))
                if os.path.isfile(path):
                    zf.write(path, os.path.relpath(path, base_path))

    # clean up
    rmtree(tempdir, ignore_errors=True)


def get_start_values(filename):
    """ Get the start values of an FMU's variables

    Parameters:
        filename    the filename of the FMU

    Returns:
        a dictionary of variables_name -> start_value
    """

    from .model_description import read_model_description
    from . import extract
    from .fmi1 import FMU1Slave, FMU1Model
    from .fmi2 import FMU2Slave, FMU2Model
    from shutil import rmtree

    unzipdir = extract(filename)

    model_description = read_model_description(filename, validate=False)

    if model_description.coSimulation is not None:
        implementation = model_description.coSimulation
    else:
        implementation = model_description.modelExchange

    # instantiate and initialize the FMU
    fmu_kwargs = {
        'guid': model_description.guid,
        'modelIdentifier': implementation.modelIdentifier,
        'unzipDirectory': unzipdir,
    }

    if model_description.fmiVersion == '1.0':
        if model_description.coSimulation is not None:
            fmu = FMU1Slave(**fmu_kwargs)
        else:
            fmu = FMU1Model(**fmu_kwargs)
        fmu.instantiate()
        fmu.initialize()
    else:
        if model_description.coSimulation is not None:
            fmu = FMU2Slave(**fmu_kwargs)
        else:
            fmu = FMU2Model(**fmu_kwargs)
        fmu.instantiate()
        fmu.enterInitializationMode()
        fmu.exitInitializationMode()

    # read the start values
    start_values = {}

    for variable in model_description.modelVariables:
        try:
            vr = [variable.valueReference]

            if variable.type == 'Real':
                value = fmu.getReal(vr=vr)
                start_values[variable.name] = str(value[0])
            elif variable.type in ['Integer', 'Enumeration']:
                value = fmu.getInteger(vr=vr)
                start_values[variable.name] = str(value[0])
            elif variable.type == 'Boolean':
                value = fmu.getBoolean(vr=vr)
                start_values[variable.name] = 'true' if value[0] != 0 else 'false'
            elif variable.type == 'String':
                value = fmu.getString(vr=vr)
                start_values[variable.name] = value[0]
        except Exception as e:
            print(e)  # do nothing

    fmu.terminate()
    fmu.freeInstance()

    # clean up
    rmtree(unzipdir, ignore_errors=True)

    return start_values


def create_cmake_project(filename, project_dir):
    """ Create a CMake project from a C code FMU

    Parameters:

        filename     filename of the C code FMU
        project_dir  existing directory for the CMake project
    """

    from zipfile import ZipFile
    from fmpy import read_model_description, extract

    model_description = read_model_description(filename)

    extract(filename, unzipdir=project_dir)

    fmpy_dir = os.path.dirname(__file__)
    source_dir = os.path.join(fmpy_dir, 'c-code')

    with open(os.path.join(source_dir, 'CMakeLists.txt'), 'r') as cmake_file:
        txt = cmake_file.read()

    definitions = []

    if model_description.coSimulation is not None:
        definitions.append('CO_SIMULATION')

    if model_description.modelExchange is not None:
        definitions.append('MODEL_EXCHANGE')

    with ZipFile(filename, 'r') as archive:
        # don't add the current directory
        resources = list(filter(lambda n: not n.startswith('.'), archive.namelist()))

    # always add the binaries
    resources.append('binaries')

    # use the first source file set of the first build configuration
    build_configuration = model_description.buildConfigurations[0]
    source_file_set = build_configuration.sourceFileSets[0]

    sources = ['sources/' + file for file in source_file_set.sourceFiles]

    # substitute the variables
    txt = txt.replace('%MODEL_NAME%', model_description.modelName)
    txt = txt.replace('%MODEL_IDENTIFIER%', build_configuration.modelIdentifier)
    txt = txt.replace('%DEFINITIONS%', ' '.join(definitions))
    txt = txt.replace('%INCLUDE_DIRS%', '"' + source_dir.replace('\\', '/') + '"')
    txt = txt.replace('%SOURCES%', ' '.join(sources))
    txt = txt.replace('%RESOURCES%', '\n    '.join('"' + r + '"' for r in resources))

    with open(os.path.join(project_dir, 'CMakeLists.txt'), 'w') as outfile:
        outfile.write(txt)


def create_jupyter_notebook(filename, notebook_filename=None):

    import nbformat as nbf
    from fmpy import read_model_description

    if notebook_filename is None:
        notebook_filename, _ = os.path.splitext(filename)
        notebook_filename += '.ipynb'

    model_description = read_model_description(filename)

    if model_description.defaultExperiment and model_description.defaultExperiment.stopTime:
        stop_time = model_description.defaultExperiment.stopTime
    else:
        stop_time = 1

    parameters = []
    output_variables = []
    max_name = 7
    max_start = 7
    max_unit = 3
    max_output = 0

    from .simulation import _get_output_variables

    for variable in _get_output_variables(model_description):
        max_output = max(max_output, len(variable.name))
        output_variables.append((variable.name, variable.description))

    for variable in model_description.modelVariables:
        if variable.causality == 'parameter' and variable.variability in ['fixed', 'tunable']:
            name, start, unit, description = variable.name, variable.start, variable.unit, variable.description
            if variable.type == 'String':
                start = "'%s'" % start
            elif variable.type == 'Boolean':
                start = 'True' if start == 'true' else 'False'
            if unit is None and variable.declaredType is not None:
                unit = variable.declaredType.unit
            max_name = max(max_name, len(name))
            max_start = max(max_start, len(start))
            max_unit = max(max_unit, len(unit)) if unit else max_unit
            parameters.append((name, start, unit, description))

    code = "import fmpy\n"
    code += "from fmpy import *\n"
    code += "\n"
    code += "fmpy.plot_library = 'plotly'  # experimental\n"
    code += "\n"
    code += "\n"
    # use relative path if possible
    if os.path.normpath(os.path.dirname(filename)) == os.path.normpath(os.path.dirname(notebook_filename)):
        code += "filename = '%s'\n" % os.path.basename(filename)
    else:
        code += "filename = r'%s'\n" % filename
    code += "\n"
    code += "start_values = {\n"
    code += "    " + "# variable".ljust(max_name + 3) + "start".rjust(max_start + 2) + "   unit".ljust(
        max_unit + 10) + "description\n"

    for name, start, unit, description in parameters:
        code += "    " + ("'" + name + "':").ljust(max_name + 3) + " "
        if unit:
            code += ("(" + start).rjust(max_start + 1)
            code += (", '" + unit + "'),").ljust(max_unit + 6)
        else:
            code += start.rjust(max_start + 1) + "," + " " * (max_unit + 5)
        if description:
            code += "  # " + description
        code += "\n"

    code += "}\n"
    code += "\n"
    code += "output = [\n"
    for name, description in output_variables:
        code += "    " + ("'%s'," % name).ljust(max_output + 3)
        if description:
            code += "  # " + description
        code += "\n"
    code += "]\n"
    code += "\n"
    code += "result = simulate_fmu(filename, start_values=start_values, output=output, stop_time=%s)\n" % stop_time
    code += "\n"
    code += "plot_result(result)"

    nb = nbf.v4.new_notebook()

    cells = []

    if model_description.description:
        cells.append(nbf.v4.new_markdown_cell(model_description.description))

    cells.append(nbf.v4.new_code_cell(code))

    nb['cells'] = cells

    with open(notebook_filename, 'w') as f:
        nbf.write(nb, f)


def has_wsl():
    """ Check if the Windows Subsystem for Linux (WSL) is available """

    if fmpy.system != 'windows':
        return False

    import subprocess

    try:
        subprocess.run(['wsl', '--help'])
        return True
    except:
        return False


def has_wine64():
    """ Check if the Wine 64-bit is available """

    if fmpy.system != 'linux':
        return False

    import subprocess

    try:
        subprocess.run(['wine64', '--help'])
        return True
    except:
        return False


def can_simulate(platforms, remote_platform='auto'):

    from . import platform

    if remote_platform is None:  # remoting disabled

        return platform in platforms, None

    elif remote_platform == 'auto':  # auto remoting

        if platform in platforms:
            return True, None
        elif platform == 'win64' and 'win32' in platforms:
            return True, 'win32'
        elif has_wsl() and 'linux64' in platforms:
            return True, 'linux64'
        elif has_wine64() and 'win64' in platforms:
            return True, 'win64'
        else:
            return False, None

    else:  # specific remoting

        if remote_platform == 'win32' and platform == 'win64':
            return True, remote_platform
        elif remote_platform == 'win64' and has_wine64():
            return True, remote_platform
        elif remote_platform == 'linux64' and has_wsl():
            return True, remote_platform

    return False, None





from ctypes import *
from .libraries import sundials_cvode

# /* -----------------------------------------------------------------
#  * Programmer(s): David J. Gardner @ LLNL
#  * -----------------------------------------------------------------
#  * SUNDIALS Copyright Start
#  * Copyright (c) 2002-2019, Lawrence Livermore National Security
#  * and Southern Methodist University.
#  * All rights reserved.
#  *
#  * See the top-level LICENSE and NOTICE files for details.
#  *
#  * SPDX-License-Identifier: BSD-3-Clause
#  * SUNDIALS Copyright End
#  * -----------------------------------------------------------------
#  * This header file is for routines to get SUNDIALS version info
#  * -----------------------------------------------------------------*/
#
# #ifndef _SUNDIALS_VERSION_H
# #define _SUNDIALS_VERSION_H
#
# #include <sundials/sundials_config.h>
#
# #ifdef __cplusplus  /* wrapper to enable C++ usage */
# extern "C" {
# #endif
#
# /* Fill a string with SUNDIALS version information */
# SUNDIALS_EXPORT int SUNDIALSGetVersion(char *version, int len);
#
# /* Fills integers with the major, minor, and patch release version numbers and a
#    string with the release label.*/
# SUNDIALS_EXPORT int SUNDIALSGetVersionNumber(int *major, int *minor, int *patch,
#                                              char *label, int len);
SUNDIALSGetVersionNumber = getattr(sundials_cvode, 'SUNDIALSGetVersionNumber')
SUNDIALSGetVersionNumber.argtypes = [POINTER(c_int), POINTER(c_int), POINTER(c_int), c_char_p, c_int]
SUNDIALSGetVersionNumber.restype = c_int
#
# #ifdef __cplusplus
# }
# #endif
#
# #endif

